from dataclasses import dataclass

from waffle_hub.schema.base_schema import BaseSchema


@dataclass
class ObjectDetectionMetric(BaseSchema):
    mAP: float


@dataclass
class ClassificationMetric(BaseSchema):
    accuracy: float


@dataclass
class InstanceSegmentationMetric(BaseSchema):
    mAP: float


@dataclass
class TextRecognitionMetric(BaseSchema):
    accuracy: float



from .autocare_dlt import export_autocare_dlt, import_autocare_dlt
from .coco import export_coco, import_coco
from .transformers import export_transformers, import_transformers
from .yolo import export_yolo, import_yolo

__all__ = [
    "export_yolo",
    "export_coco",
    "export_transformers",
    "export_autocare_dlt",
    "import_autocare_dlt",
    "import_coco",
    "import_transformers",
    "import_yolo",
]



from waffle_hub import TaskType
from waffle_hub.schema.configs import TrainConfig

# Common
MODEL_TYPES = {
    TaskType.OBJECT_DETECTION: {
        "yolov8": {
            "n": "yolov8n.pt",
            "s": "yolov8s.pt",
            "m": "yolov8m.pt",
            "l": "yolov8l.pt",
            "x": "yolov8x.pt",
        },
        "yolov5": {
            "n": "yolov5nu.pt",
            "s": "yolov5su.pt",
            "m": "yolov5mu.pt",
            "l": "yolov5lu.pt",
            "x": "yolov5xu.pt",
            "n6": "yolov5n6u.pt",
            "s6": "yolov5s6u.pt",
            "m6": "yolov5m6u.pt",
            "l6": "yolov5l6u.pt",
            "x6": "yolov5x6u.pt",
        },
    },
    TaskType.CLASSIFICATION: {
        "yolov8": {
            "n": "yolov8n-cls.pt",
            "s": "yolov8s-cls.pt",
            "m": "yolov8m-cls.pt",
            "l": "yolov8l-cls.pt",
            "x": "yolov8x-cls.pt",
        }
    },
    TaskType.INSTANCE_SEGMENTATION: {
        "yolov8": {
            "n": "yolov8n-seg.pt",
            "s": "yolov8s-seg.pt",
            "m": "yolov8m-seg.pt",
            "l": "yolov8l-seg.pt",
            "x": "yolov8x-seg.pt",
        }
    },
    # "keypoint_detection": {"yolov8": list("nsmlx")},
}

# Backend Specifics
TASK_MAP = {
    TaskType.OBJECT_DETECTION: "detect",
    TaskType.CLASSIFICATION: "classify",
    TaskType.INSTANCE_SEGMENTATION: "segment"
    # "keypoint_detection": "pose"
}

DEFAULT_PARAMS = {
    TaskType.OBJECT_DETECTION: {
        "yolov8": {
            "n": TrainConfig(
                epochs=50,
                image_size=[640, 640],
                learning_rate=0.01,
                letter_box=True,
                batch_size=64,
            ),
            "s": TrainConfig(
                epochs=50,
                image_size=[640, 640],
                learning_rate=0.01,
                letter_box=True,
                batch_size=32,
            ),
            "m": TrainConfig(
                epochs=50,
                image_size=[640, 640],
                learning_rate=0.01,
                letter_box=True,
                batch_size=16,
            ),
            "l": TrainConfig(
                epochs=50,
                image_size=[640, 640],
                learning_rate=0.01,
                letter_box=True,
                batch_size=8,
            ),
            "x": TrainConfig(
                epochs=50,
                image_size=[640, 640],
                learning_rate=0.01,
                letter_box=True,
                batch_size=8,
            ),
        },
        "yolov5": {
            "n": TrainConfig(
                epochs=50,
                image_size=[640, 640],
                learning_rate=0.01,
                letter_box=True,
                batch_size=64,
            ),
            "s": TrainConfig(
                epochs=50,
                image_size=[640, 640],
                learning_rate=0.01,
                letter_box=True,
                batch_size=32,
            ),
            "m": TrainConfig(
                epochs=50,
                image_size=[640, 640],
                learning_rate=0.01,
                letter_box=True,
                batch_size=16,
            ),
            "l": TrainConfig(
                epochs=50,
                image_size=[640, 640],
                learning_rate=0.01,
                letter_box=True,
                batch_size=8,
            ),
            "x": TrainConfig(
                epochs=50,
                image_size=[640, 640],
                learning_rate=0.01,
                letter_box=True,
                batch_size=8,
            ),
            "n6": TrainConfig(
                epochs=50,
                image_size=[1280, 1280],
                learning_rate=0.01,
                letter_box=True,
                batch_size=32,
            ),
            "s6": TrainConfig(
                epochs=50,
                image_size=[1280, 1280],
                learning_rate=0.01,
                letter_box=True,
                batch_size=16,
            ),
            "m6": TrainConfig(
                epochs=50,
                image_size=[1280, 1280],
                learning_rate=0.01,
                letter_box=True,
                batch_size=8,
            ),
            "l6": TrainConfig(
                epochs=50,
                image_size=[1280, 1280],
                learning_rate=0.01,
                letter_box=True,
                batch_size=4,
            ),
            "x6": TrainConfig(
                epochs=50,
                image_size=[1280, 1280],
                learning_rate=0.01,
                letter_box=True,
                batch_size=2,
            ),
        },
    },
    TaskType.CLASSIFICATION: {
        "yolov8": {
            "n": TrainConfig(
                epochs=50,
                image_size=[224, 224],
                learning_rate=0.01,
                letter_box=False,
                batch_size=512,
            ),
            "s": TrainConfig(
                epochs=50,
                image_size=[224, 224],
                learning_rate=0.01,
                letter_box=False,
                batch_size=256,
            ),
            "m": TrainConfig(
                epochs=50,
                image_size=[224, 224],
                learning_rate=0.01,
                letter_box=False,
                batch_size=128,
            ),
            "l": TrainConfig(
                epochs=50,
                image_size=[224, 224],
                learning_rate=0.01,
                letter_box=False,
                batch_size=64,
            ),
            "x": TrainConfig(
                epochs=50,
                image_size=[224, 224],
                learning_rate=0.01,
                letter_box=False,
                batch_size=64,
            ),
        }
    },
    TaskType.INSTANCE_SEGMENTATION: {
        "yolov8": {
            "n": TrainConfig(
                epochs=50,
                image_size=[640, 640],
                learning_rate=0.01,
                letter_box=True,
                batch_size=32,
            ),
            "s": TrainConfig(
                epochs=50,
                image_size=[640, 640],
                learning_rate=0.01,
                letter_box=True,
                batch_size=16,
            ),
            "m": TrainConfig(
                epochs=50,
                image_size=[640, 640],
                learning_rate=0.01,
                letter_box=True,
                batch_size=8,
            ),
            "l": TrainConfig(
                epochs=50,
                image_size=[640, 640],
                learning_rate=0.01,
                letter_box=True,
                batch_size=4,
            ),
            "x": TrainConfig(
                epochs=50,
                image_size=[640, 640],
                learning_rate=0.01,
                letter_box=True,
                batch_size=4,
            ),
        }
    },
}



"""
Pseudo code for Training

Use Case 1.
# start train
train(..., hold=True)
# end train

Use Case 2.
# start train in another thread
callback = train(..., hold=False)

# wait for train to finish
while callback.end():
    print(callback.get_progress())

# end train

Sample code for Training
def train(hold: bool=True) -> Union[None, TrainCallback]:

    if hold:
        # train and return
        return
    else:
        # train in another thread and return callback
        return callback

Sample code for Inference
def inference(hold: bool=True) -> Union[None, InferenceCallback]:

    if hold:
        # inference and return
        return
    else:
        # inference in another thread and return callback
        return callback
"""
import threading
import time
import warnings


class ThreadProgressCallback:
    def __init__(self, total_steps: int):
        self._total_steps = total_steps

        self._thread = None
        self._finished = False
        self._failed = False
        self._progress = 0
        self._start_time = time.time()

    def get_progress(self) -> float:
        """Get the progress of the task. (0 ~ 1)"""
        return self._progress

    def is_finished(self) -> bool:
        """Check if the task has finished."""
        return self._finished

    def is_failed(self) -> bool:
        """Check if the task has failed."""
        return self._failed

    def set_failed(self):
        """Set the task as failed."""
        self._failed = True

    def get_remaining_time(self) -> float:
        """Get the remaining time of the task. (seconds)"""
        elapsed = time.time() - self._start_time
        if self._progress == 0:
            return float("inf")
        return (elapsed / self._progress) - elapsed

    def update(self, step: int):
        """Update the progress of the task. (0 ~ total_steps)"""
        if self._finished:
            warnings.warn("Callback has already ended")
        elif step >= self._total_steps:
            self._finished = True
            self._progress = step / self._total_steps
        else:
            self._progress = step / self._total_steps

    def force_finish(self):
        """Force the task to end."""
        self._finished = True
        self._progress = 1.0

    def register_thread(self, thread: threading.Thread):
        """Register the thread that is running the task."""
        self._thread = thread

    def start(self):
        """Start the thread that is running the task."""
        if self._thread is not None:
            self._thread.start()

    def join(self):
        """Wait for the thread that is running the task to end."""
        if self._thread is not None:
            self._thread.join()


class TrainCallback(ThreadProgressCallback):
    def __init__(self, total_steps: int, get_metric_func):
        super().__init__(total_steps)

        self._get_metric_func = get_metric_func

    def get_progress(self) -> float:
        """Get the progress of the task. (0 ~ 1)"""
        if not self._finished:
            metrics = self._get_metric_func()
            if len(metrics) == 0:
                return 0
            self.update(len(metrics))
            self._progress = len(metrics) / self._total_steps
            return self._progress
        else:
            return 1.0


class EvaluateCallback(ThreadProgressCallback):
    def __init__(self, total_steps: int):
        super().__init__(total_steps)


class InferenceCallback(ThreadProgressCallback):
    def __init__(self, total_steps: int):
        super().__init__(total_steps)


class ExportCallback(ThreadProgressCallback):
    def __init__(self, total_steps: int):
        super().__init__(total_steps)



from typing import Union

from waffle_utils.utils import type_validator

from waffle_hub import TaskType

from .base_field import BaseField


class Category(BaseField):
    def __init__(
        self,
        # required
        category_id: int,
        name: str,
        supercategory: str = None,
        # for keypoint detection
        keypoints: list[str] = None,
        skeleton: list[list[int]] = None,
        #
        task: Union[str, TaskType] = None,
    ):

        self.category_id = category_id
        self.supercategory = supercategory
        self.name = name
        self.keypoints = keypoints
        self.skeleton = skeleton

        self.task = task

    # properties
    @property
    def category_id(self):
        return self.__category_id

    @category_id.setter
    @type_validator(int)
    def category_id(self, v):
        if v and v < 1:
            raise ValueError("id should be greater than 0.")
        self.__category_id = v

    @property
    def supercategory(self):
        return self.__supercategory

    @supercategory.setter
    def supercategory(self, v):
        self.__supercategory = v if v else "object"

    @property
    def name(self):
        return self.__name

    @name.setter
    @type_validator(str)
    def name(self, v):
        self.__name = v

    @property
    def keypoints(self):
        return self.__keypoints

    @keypoints.setter
    @type_validator(list)
    def keypoints(self, v):
        self.__keypoints = v

    @property
    def skeleton(self):
        return self.__skeleton

    @skeleton.setter
    @type_validator(list)
    def skeleton(self, v):
        self.__skeleton = v

    @property
    def task(self):
        return self.__task

    @task.setter
    def task(self, v):
        if v is not None and v not in TaskType:
            raise ValueError(f"Invalid task type: {v}" f"Available task types: {list(TaskType)}")
        self.__task = str(v).upper()

    # factories
    @classmethod
    def new(
        cls,
        category_id: int,
        name: str,
        supercategory: str = None,
        keypoints: list[str] = None,
        skeleton: list[list[int]] = None,
        task: Union[str, TaskType] = None,
        **kwargs,
    ) -> "Category":
        """Category Format

        Args:
            category_id (int): category id. natural number.
            name (str): category name.
            supercategory (str): supercategory name.
            keypoints (list[str]): category name.
            skeleton (list[list[int]]): skeleton edges.
            task (Union[str, TaskType], optional): task type. Default to None.

        Returns:
            Category: category class
        """
        return (getattr(cls, task.lower()) if task else cls)(
            category_id=category_id,
            name=name,
            supercategory=supercategory,
            keypoints=keypoints,
            skeleton=skeleton,
            task=task,
        )

    @classmethod
    def classification(
        cls, category_id: int, name: str, supercategory: str = None, **kwargs
    ) -> "Category":
        """Classification Category Format

        Args:
            category_id (int): category id. natural number.
            name (str): category name.
            supercategory (str): supercategory name.

        Returns:
            Category: category class
        """
        return cls(
            category_id=category_id,
            name=name,
            supercategory=supercategory,
            task=TaskType.CLASSIFICATION,
        )

    @classmethod
    def object_detection(
        cls, category_id: int, name: str, supercategory: str = None, **kwargs
    ) -> "Category":
        """Object Detection Category Format

        Args:
            category_id (int): category id. natural number.
            name (str): category name.
            supercategory (str): supercategory name.

        Returns:
            Category: category class
        """
        return cls(
            category_id=category_id,
            name=name,
            supercategory=supercategory,
            task=TaskType.OBJECT_DETECTION,
        )

    @classmethod
    def semantic_segmentation(
        cls, category_id: int, name: str, supercategory: str = None, **kwargs
    ) -> "Category":
        """Segmentation Category Format

        Args:
            category_id (int): category id. natural number.
            name (str): category name.
            supercategory (str): supercategory name.

        Returns:
            Category: category class
        """
        return cls(
            category_id=category_id,
            name=name,
            supercategory=supercategory,
            task=TaskType.SEMANTIC_SEGMENTATION,
        )

    @classmethod
    def instance_segmentation(
        cls, category_id: int, name: str, supercategory: str = None, **kwargs
    ) -> "Category":
        """Instance Category Format

        Args:
            category_id (int): category id. natural number.
            name (str): category name.
            supercategory (str): supercategory name.

        Returns:
            Category: category class
        """
        return cls(
            category_id=category_id,
            name=name,
            supercategory=supercategory,
            task=TaskType.INSTANCE_SEGMENTATION,
        )

    @classmethod
    def keypoint_detection(
        cls,
        category_id: int,
        name: str,
        keypoints: list[str],
        skeleton: list[list[int]],
        supercategory: str = None,
        **kwargs,
    ) -> "Category":
        """Keypoint Detection Category Format

        Args:
            category_id (int): category id. natural number.
            name (str): category name.
            keypoints (list[str]): category name.
            skeleton (list[list[int]]): skeleton edges.
            supercategory (str): supercategory name.

        Returns:
            Category: category class
        """
        return cls(
            category_id=category_id,
            name=name,
            supercategory=supercategory,
            keypoints=keypoints,
            skeleton=skeleton,
            task=TaskType.KEYPOINT_DETECTION,
        )

    @classmethod
    def text_recognition(
        cls, category_id: int, name: str, supercategory: str = None, **kwargs
    ) -> "Category":
        """Text Recognition Category Format

        Args:
            category_id (int): category id. natural number.
            name (str): category name.
            supercategory (str): supercategory name.

        Returns:
            Category: category class
        """
        return cls(
            category_id=category_id,
            name=name,
            supercategory=supercategory,
            task=TaskType.TEXT_RECOGNITION,
        )

    def to_dict(self) -> dict:
        """Get Dictionary of Category

        Returns:
            dict: annotation dictionary.
        """

        cat = {
            "category_id": self.category_id,
            "supercategory": self.supercategory,
            "name": self.name,
        }

        if self.keypoints is not None:
            cat["keypoints"] = self.keypoints
        if self.skeleton is not None:
            cat["skeleton"] = self.skeleton

        return cat



def get_model_config(
    model_type: str,
    model_size: str,
    categories: list[str],
    seed: int,
    learning_rate: float,
    letter_box: bool,
    epochs: int,
):
    if model_type == "YOLOv5":
        return {
            "task": model_type,
            "model": {
                "model_size": model_size,
                "backbone": {"name": "YOLOv5Backbone", "focus": False},
                "neck": {"name": "YOLOv5Neck"},
                "head": {
                    "name": "YOLOv5Head",
                    "anchors": [
                        [10, 13, 16, 30, 33, 23],
                        [30, 61, 62, 45, 59, 119],
                        [116, 90, 156, 198, 373, 326],
                    ],
                },
            },
            "loss": {
                "total_loss": {
                    "name": "YoloLoss",
                    "params": {
                        "hyp": {
                            "box": 0.05,
                            "cls": 0.3,
                            "cls_pw": 1.0,
                            "obj": 0.7,
                            "obj_pw": 1.0,
                            "anchor_t": 4.0,
                            "fl_alpha": -1,
                            "fl_gamma": 0.0,
                            "label_smoothing": 0.0,
                        }
                    },
                }
            },
            "optim": {
                "name": "SGD",
                "lr": learning_rate,
                "momentum": 0.937,
                "weight_decay": 0.0005,
            },
            "lr_cfg": {
                "type": "cosine",
                "warmup": True,
                "warmup_epochs": 0.5,
                "lrf": 0.1,
            },
            "ema_cfg": {"burn_in_epoch": 1},
            "max_epoch": epochs,
            "nms_thresh": 0.65,
            "min_score": 0.25,
            "detections_per_img": 300,
            "seed": seed,
            "categories": categories,
            "num_classes": len(categories),
            "classes": categories,
            "letter_box": letter_box,
        }

    elif model_type == "Classifier":
        if model_size == "s":
            backbone = "resnet18"
        elif model_size == "m":
            backbone = "resnet34"
        elif model_size == "l":
            backbone = "resnet50"

        return {
            "task": model_type,
            "model": {
                "backbone": {"name": backbone},
                "neck": {"name": "Identity"},
                "head": {"name": "ClassificationHead", "in_channels": 512},
            },
            "loss": {
                "cls_loss": {
                    "name": "CrossEntropyLoss",
                    "params": {"ignore_index": -1},
                }
            },
            "optim": {
                "name": "SGD",
                "lr": learning_rate,
                "momentum": 0.9,
                "weight_decay": 0.0001,
            },
            "lr_cfg": {"type": "cosine"},
            "use_model_ema": True,
            "max_epoch": epochs,
            "seed": seed,
            "classes": categories,
            "num_classes": sum([len(list(category.values())[0]) for category in categories]),
        }

    elif model_type == "TextRecognition":
        if model_size == "s":
            backbone = "resnet18"
        elif model_size == "m":
            backbone = "resnet34"
        elif model_size == "l":
            backbone = "resnet50"

        return {
            "task": model_type,
            "model": {
                "Transformation": {"name": None},
                "FeatureExtraction": {"name": backbone, "feature_index": 3, "output_size": 256},
                "SequenceModeling": {"name": "BiLSTM", "input_size": 256, "hidden_size": 256},
                "Prediction": {"name": "CTC", "input_size": 256},
                "max_string_length": 50,
            },
            "loss": {"str_loss": {"name": "STRCTCLoss", "params": {}}},
            "optim": {"name": "Adam", "lr": learning_rate},
            "lr_cfg": {"type": "cosine"},
            "use_model_ema": True,
            "max_epoch": epochs,
            "seed": seed,
            "classes": categories,
            "num_classes": len(categories),
        }
    elif model_type == "LicencePlateRecognition":
        if model_size == "s":
            backbone = "resnet18"
        elif model_size == "m":
            backbone = "resnet34"
        elif model_size == "l":
            backbone = "resnet50"

        return {
            "task": model_type,
            "model": {
                "Transformation": {"name": None},
                "FeatureExtraction": {"name": backbone, "feature_index": 3, "output_size": 256},
                "SequenceModeling": {"name": "BiLSTM", "input_size": 256, "hidden_size": 256},
                "Prediction": {"name": "CTC", "input_size": 256},
                "max_string_length": 10,
            },
            "loss": {"str_loss": {"name": "LPRLoss", "params": {}}},
            "optim": {"name": "Adam", "lr": learning_rate},
            "lr_cfg": {"type": "cosine", "warmup": True, "warmup_epochs": 1},
            "ema_cfg": {"burn_in_epoch": 5},
            "max_epoch": epochs,
            "seed": seed,
            "classes": categories,
            "num_classes": len(categories),
        }
    else:
        raise NotImplementedError(f"Model type {model_type} not implemented.")



import logging
import warnings
from pathlib import Path
from typing import Union

import tqdm
from pycocotools.coco import COCO
from waffle_utils.file import io

from waffle_hub import TaskType
from waffle_hub.schema.fields import Annotation, Category, Image
from waffle_hub.utils.conversion import convert_rle_to_polygon


def _export_coco(
    self,
    export_dir: Path,
    train_ids: list,
    val_ids: list,
    test_ids: list,
    unlabeled_ids: list,
):
    """Export dataset to COCO format

    Args:
        export_dir (Path): Path to export directory
        train_ids (list): List of train ids
        val_ids (list): List of validation ids
        test_ids (list): List of test ids
        unlabeled_ids (list): List of unlabeled ids
    """
    io.make_directory(export_dir)

    image_dir = export_dir / "images"

    for split, image_ids in zip(
        ["train", "val", "test", "unlabeled"],
        [train_ids, val_ids, test_ids, unlabeled_ids],
    ):
        if len(image_ids) == 0:
            continue

        coco = {
            "categories": [
                {
                    "id": category.category_id,
                    "name": category.name,
                    "supercategory": category.supercategory,
                }
                for category in self.get_categories()
            ],
            "images": [],
            "annotations": [],
        }

        for image in self.get_images(image_ids):
            image_path = self.raw_image_dir / image.file_name
            image_dst_path = image_dir / image.file_name
            io.copy_file(image_path, image_dst_path, create_directory=True)

            d = image.to_dict()
            image_id = d.pop("image_id")
            coco["images"].append({"id": image_id, **d})

            annotations = self.get_annotations(image_id)
            for annotation in annotations:
                d = annotation.to_dict()
                if d.get("segmentation", None):
                    if isinstance(d["segmentation"], dict):
                        d["segmentation"] = convert_rle_to_polygon(d["segmentation"])
                annotation_id = d.pop("annotation_id")
                coco["annotations"].append({"id": annotation_id, **d})

        io.save_json(coco, export_dir / f"{split}.json", create_directory=True)


def export_coco(self, export_dir: Union[str, Path]) -> str:
    """Export dataset to COCO format

    Args:
        export_dir (str): Path to export directory

    Returns:
        str: Path to export directory
    """
    export_dir = Path(export_dir)

    train_ids, val_ids, test_ids, _ = self.get_split_ids()

    if self.task == TaskType.CLASSIFICATION:
        _export_coco(self, export_dir, train_ids, val_ids, test_ids, [])
    elif self.task == TaskType.OBJECT_DETECTION:
        _export_coco(self, export_dir, train_ids, val_ids, test_ids, [])
    elif self.task == TaskType.INSTANCE_SEGMENTATION:
        _export_coco(self, export_dir, train_ids, val_ids, test_ids, [])
    else:
        raise ValueError(f"Unsupported task type: {self.task}")

    return str(export_dir)


def import_coco(self, coco_files: list[str], coco_root_dirs: list[str]):
    """
    Import coco dataset

    Args:
        coco_files (list[str]): List of coco annotation files
        coco_root_dirs (list[str]): List of coco root directories
    """
    if len(coco_files) == 1:
        set_names = [None]
    elif len(coco_files) == 2:
        set_names = ["train", "val"]
    elif len(coco_files) == 3:
        set_names = ["train", "val", "test"]
    else:
        raise ValueError("coco_file should have 1, 2, or 3 files.")

    cocos = [COCO(coco_file) for coco_file in coco_files]

    # categories should be same between coco files
    categories = cocos[0].loadCats(cocos[0].getCatIds())
    for coco in cocos[1:]:
        if categories != coco.loadCats(coco.getCatIds()):
            raise ValueError("categories should be same between coco files.")

    coco_cat_id_to_waffle_cat_id = {}
    for i, category in enumerate(categories, start=1):
        coco_category_id = category.pop("id")
        coco_cat_id_to_waffle_cat_id[coco_category_id] = i
        self.add_categories([Category.from_dict({**category, "category_id": i}, task=self.task)])

    # import coco dataset
    total_length = sum([len(coco.getImgIds()) for coco in cocos])
    logging.info(f"Importing coco dataset. Total length: {total_length}")
    pgbar = tqdm.tqdm(total=total_length, desc="Importing coco dataset")

    image_id = 1
    annotation_id = 1

    # parse coco annotation file
    for coco, coco_root_dir, set_name in tqdm.tqdm(zip(cocos, coco_root_dirs, set_names)):

        image_ids = []
        for coco_image_id, annotation_dicts in coco.imgToAnns.items():
            if len(annotation_dicts) == 0:
                warnings.warn(f"image_id {coco_image_id} has no annotations.")
                continue

            image_dict = coco.loadImgs(coco_image_id)[0]
            image_dict.pop("id")

            file_name = image_dict.pop("file_name")
            image_path = Path(coco_root_dir) / file_name
            if not image_path.exists():
                raise FileNotFoundError(f"{image_path} does not exist.")

            if set_name:
                file_name = f"{set_name}/{file_name}"

            self.add_images(
                [Image.from_dict({**image_dict, "image_id": image_id, "file_name": file_name})]
            )
            io.copy_file(image_path, self.raw_image_dir / file_name, create_directory=True)

            for annotation_dict in annotation_dicts:
                annotation_dict.pop("id")
                self.add_annotations(
                    [
                        Annotation.from_dict(
                            {
                                **annotation_dict,
                                "image_id": image_id,
                                "annotation_id": annotation_id,
                                "category_id": coco_cat_id_to_waffle_cat_id[
                                    annotation_dict["category_id"]
                                ],
                            },
                            task=self.task,
                        )
                    ]
                )
                annotation_id += 1

            image_ids.append(image_id)
            image_id += 1
            pgbar.update(1)

        if set_name:
            io.save_json(image_ids, self.set_dir / f"{set_name}.json", create_directory=True)

    pgbar.close()



import copy
import logging
import os
import random
import shutil
import time
import warnings
from collections import Counter, OrderedDict, defaultdict
from functools import cached_property
from pathlib import Path
from tempfile import mkdtemp
from typing import Union

import PIL.Image
import tqdm
from waffle_utils.file import io, network
from waffle_utils.image.io import load_image, save_image
from waffle_utils.log import datetime_now
from waffle_utils.utils import type_validator

from waffle_hub import EXPORT_MAP, DataType, SplitMethod, TaskType
from waffle_hub.dataset.adapter import (
    export_autocare_dlt,
    export_coco,
    export_transformers,
    export_yolo,
    import_autocare_dlt,
    import_coco,
    import_transformers,
    import_yolo,
)
from waffle_hub.schema import Annotation, Category, DatasetInfo, Image
from waffle_hub.utils.draw import draw_results

logger = logging.getLogger(__name__)


class Dataset:
    DEFAULT_DATASET_ROOT_DIR = Path("./datasets")
    DATASET_INFO_FILE_NAME = Path("info.yaml")

    RAW_IMAGE_DIR = Path("raw")
    IMAGE_DIR = Path("images")
    ANNOTATION_DIR = Path("annotations")
    CATEGORY_DIR = Path("categories")
    PREDICTION_DIR = Path("predictions")
    EXPORT_DIR = Path("exports")
    SET_DIR = Path("sets")
    DRAW_DIR = Path("draws")

    TRAIN_SET_FILE_NAME = Path("train.json")
    VAL_SET_FILE_NAME = Path("val.json")
    TEST_SET_FILE_NAME = Path("test.json")
    UNLABELED_SET_FILE_NAME = Path("unlabeled.json")

    MINIMUM_TRAINABLE_IMAGE_NUM_PER_CATEGORY = 3

    def __init__(
        self,
        name: str,
        task: Union[str, TaskType],
        categories: list[Union[str, int, float, dict, Category]] = None,
        created: str = None,
        root_dir: str = None,
    ):
        self.name = name
        self.task = task
        self.created = created

        self.root_dir = root_dir

        if not self.initialized():
            self.initialize()
            self.set_categories(categories)
            self.save_dataset_info()
        else:  # for backward compatibility
            self.save_dataset_info()

    def __repr__(self):
        return self.get_dataset_info().__repr__()

    # properties
    @property
    def name(self):
        return self.__name

    @name.setter
    @type_validator(str)
    def name(self, v):
        self.__name = v

    @property
    def task(self):
        return self.__task

    @task.setter
    def task(self, v):
        v = str(v).upper()
        if v not in TaskType:
            raise ValueError(f"Invalid task type: {v}" f"Available task types: {list(TaskType)}")
        self.__task = v

    @property
    def categories(self) -> list[Category]:
        return self.get_categories()

    def set_categories(self, v):
        if v is None or len(v) == 0:
            v = []
        elif isinstance(v[0], dict):
            v = [
                getattr(Category, self.task.lower())(
                    **{
                        **category,
                        "category_id": category.get("category_id", i),
                    }
                )
                for i, category in enumerate(v, start=1)
            ]
        elif isinstance(v[0], (str, int, float)):
            v = [
                getattr(Category, self.task.lower())(
                    category_id=i,
                    supercategory="object",
                    name=str(category),
                )
                for i, category in enumerate(v, start=1)
            ]
        elif isinstance(v[0], Category):
            pass

        self.add_categories(v)

    def extract_by_image_ids(
        self, name: str, image_ids: list[int], root_dir: str = None
    ) -> "Dataset":
        """
        Extract a new dataset by image ids

        Args:
            name (str): Name of the new dataset
            image_ids (list[int]): Image ids to extract
            root_dir (str, optional): Root directory of the new dataset. Defaults to None.

        Returns:
            Dataset: Extracted dataset

        """
        ds = Dataset.new(
            name=name,
            task=self.task,
            root_dir=root_dir,
        )

        try:
            ds.add_categories(self.get_categories())
            for image in self.get_images(image_ids):
                annotations = self.get_annotations(image.image_id)
                io.copy_file(
                    self.raw_image_dir / image.file_name, ds.raw_image_dir / image.file_name
                )
                ds.add_images([image])
                ds.add_annotations(annotations)

        except Exception as e:
            ds.delete()
            raise e

        return ds

    def extract_by_categories(
        self, name: str, category_ids: list[int], root_dir: str = None
    ) -> "Dataset":
        """
        Extract a new dataset by categories

        Args:
            name (str): Name of the new dataset
            category_ids (list[int]): Category IDs to extract
            root_dir (str, optional): Root directory of the new dataset. Defaults to None.

        Returns (Dataset): New dataset
        """
        ds = Dataset.new(
            name=name,
            task=self.task,
            root_dir=root_dir,
        )
        try:
            category_old2new = {}
            for new_category_id, category_id in enumerate(category_ids, start=1):
                category_old2new[category_id] = new_category_id
                categories = self.get_categories([category_id])
                for category in categories:
                    category.category_id = new_category_id
                ds.add_categories(categories)

            for image in self.get_images():
                annotations = list(
                    filter(
                        lambda ann: ann.category_id in category_ids,
                        self.get_annotations(image.image_id),
                    )
                )
                for annotation in annotations:
                    annotation.category_id = category_old2new[annotation.category_id]

                if annotations:
                    io.copy_file(
                        self.raw_image_dir / image.file_name, ds.raw_image_dir / image.file_name
                    )
                    ds.add_images([image])
                    ds.add_annotations(annotations)

        except Exception as e:
            ds.delete()
            raise e

        return ds

    @property
    def created(self):
        return self.__created

    @created.setter
    def created(self, v):
        self.__created = v or datetime_now()

    @property
    def root_dir(self) -> Path:
        return self.__root_dir

    @root_dir.setter
    @type_validator(Path, strict=False)
    def root_dir(self, v):
        self.__root_dir = Dataset.parse_root_dir(v)
        logger.info(f"Dataset root directory: {self.__root_dir}")

    @classmethod
    def parse_root_dir(cls, v):
        if v:
            return Path(v)
        elif os.getenv("WAFFLE_DATASET_ROOT_DIR", None):
            return Path(os.getenv("WAFFLE_DATASET_ROOT_DIR"))
        else:
            return Dataset.DEFAULT_DATASET_ROOT_DIR

    # cached properties
    @cached_property
    def dataset_dir(self) -> Path:
        return self.root_dir / self.name

    @cached_property
    def dataset_info_file(self) -> Path:
        return self.dataset_dir / Dataset.DATASET_INFO_FILE_NAME

    @cached_property
    def raw_image_dir(self) -> Path:
        return self.dataset_dir / Dataset.RAW_IMAGE_DIR

    @cached_property
    def image_dir(self) -> Path:
        return self.dataset_dir / Dataset.IMAGE_DIR

    @cached_property
    def annotation_dir(self) -> Path:
        return self.dataset_dir / Dataset.ANNOTATION_DIR

    @cached_property
    def prediction_dir(self) -> Path:
        return self.dataset_dir / Dataset.PREDICTION_DIR

    @cached_property
    def category_dir(self) -> Path:
        return self.dataset_dir / Dataset.CATEGORY_DIR

    @cached_property
    def export_dir(self) -> Path:
        return self.dataset_dir / Dataset.EXPORT_DIR

    @cached_property
    def set_dir(self) -> Path:
        return self.dataset_dir / Dataset.SET_DIR

    @cached_property
    def draw_dir(self) -> Path:
        return self.dataset_dir / Dataset.DRAW_DIR

    @cached_property
    def train_set_file(self) -> Path:
        return self.set_dir / Dataset.TRAIN_SET_FILE_NAME

    @cached_property
    def val_set_file(self) -> Path:
        return self.set_dir / Dataset.VAL_SET_FILE_NAME

    @cached_property
    def test_set_file(self) -> Path:
        return self.set_dir / Dataset.TEST_SET_FILE_NAME

    @cached_property
    def unlabeled_set_file(self) -> Path:
        return self.set_dir / Dataset.UNLABELED_SET_FILE_NAME

    # dataset indexes
    @property
    def image_dict(self) -> dict[int, Image]:
        if not hasattr(self, "_image_dict"):
            self.create_index()
        return self._image_dict

    @property
    def unlabeled_image_dict(self) -> dict[int, Image]:
        if not hasattr(self, "_unlabeled_image_dict"):
            self.create_index()
        return self._unlabeled_image_dict

    @property
    def annotation_dict(self) -> dict[int, Annotation]:
        if not hasattr(self, "_annotation_dict"):
            self.create_index()
        return self._annotation_dict

    @property
    def prediction_dict(self) -> dict[int, Annotation]:
        if not hasattr(self, "_prediction_dict"):
            self.create_index()
        return self._prediction_dict

    @property
    def category_dict(self) -> dict[int, Category]:
        if not hasattr(self, "_category_dict"):
            self.create_index()
        return self._category_dict

    @property
    def image_to_annotations(self) -> dict[int, list[Annotation]]:
        if not hasattr(self, "_image_to_annotations"):
            self.create_index()
        return self._image_to_annotations

    @property
    def image_to_predictions(self) -> dict[int, list[Annotation]]:
        if not hasattr(self, "_image_to_predictions"):
            self.create_index()
        return self._image_to_predictions

    @property
    def annotation_to_image(self) -> dict[int, Image]:
        if not hasattr(self, "_annotation_to_image"):
            self.create_index()
        return self._annotation_to_image

    @property
    def prediction_to_image(self) -> dict[int, Image]:
        if not hasattr(self, "_prediction_to_image"):
            self.create_index()
        return self._prediction_to_image

    @property
    def category_to_images(self) -> dict[int, list[Image]]:
        if not hasattr(self, "_category_to_images"):
            self.create_index()
        return self._category_to_images

    @property
    def category_to_unique_images(self) -> dict[int, list[Image]]:
        if not hasattr(self, "_category_to_unique_images"):
            self.create_index()
        return self._category_to_unique_images

    @property
    def category_name_to_category(self) -> dict[str, Category]:
        if not hasattr(self, "_category_name_to_category"):
            self.create_index()
        return self._category_name_to_category

    @property
    def category_to_annotations(self) -> dict[int, list[Annotation]]:
        if not hasattr(self, "_category_to_annotations"):
            self.create_index()
        return self._category_to_annotations

    @property
    def category_to_predictions(self) -> dict[int, list[Annotation]]:
        if not hasattr(self, "_category_to_predictions"):
            self.create_index()
        return self._category_to_predictions

    def get_category_names(self) -> list[str]:
        return [category.name for category in self.categories]

    # factories
    @classmethod
    def new(
        cls,
        name: str,
        task: str,
        categories: list[Union[str, int, float, dict, Category]] = None,
        root_dir: str = None,
    ) -> "Dataset":
        """
        Create New Dataset.
        This method creates a new dataset directory and initialize dataset info file.
        If you have other types of data, you can use from_* methods to create a dataset.

        Args:
            name (str): Dataset name
            task (str): Dataset task
            categories (list[Union[str, int, float, dict, Category]]): Dataset categories
            root_dir (str, optional): Dataset root directory. Defaults to None.

        Raises:
            FileExistsError: if dataset name already exists

        Examples:
            >>> ds = Dataset.new("my_dataset", "CLASSIFICATION")
            >>> ds.name
            'my_dataset'  # dataset name
            >>> ds.task  # dataset task
            'CLASSIFICATION'

        Returns:
            Dataset: Dataset Class
        """
        root_dir = Dataset.parse_root_dir(root_dir)

        if name in Dataset.get_dataset_list(root_dir):
            raise FileExistsError(f"Dataset {name} already exists.")

        try:
            return cls(name=name, task=task, categories=categories, root_dir=root_dir)
        except Exception as e:
            if (root_dir / name).exists():
                io.remove_directory(root_dir / name)
            raise e

    @classmethod
    def clone(
        cls,
        src_name: str,
        name: str,
        src_root_dir: str = None,
        root_dir: str = None,
    ) -> "Dataset":
        """
        Clone Existing Dataset.
        This method clones an existing dataset.

        Args:
            src_name (str):
                Dataset name to clone.
                It should be Waffle Created Dataset.
            name (str): New Dataset name
            src_root_dir (str, optional): Source Dataset root directory. Defaults to None.
            root_dir (str, optional): New Dataset root directory. Defaults to None.

        Raises:
            FileNotFoundError: if source dataset does not exist.
            FileExistsError: if new dataset name already exist.

        Examples:
            >>> ds = Dataset.clone("my_dataset", "my_dataset_clone")
            >>> ds.name
            'my_dataset_clone'  # cloned dataset name
            >>> ds.task
            'CLASSIFICATION'   # original dataset task

        Returns:
            Dataset: Dataset Class
        """
        root_dir = Dataset.parse_root_dir(root_dir)

        try:
            src_ds = Dataset.load(src_name, src_root_dir)

            ds = Dataset.new(name=name, task=src_ds.task, root_dir=root_dir)
            io.copy_files_to_directory(src_ds.dataset_dir, ds.dataset_dir, create_directory=True)
            ds.save_dataset_info()

            ds.create_index()
            return ds
        except Exception as e:
            if (root_dir / name).exists():
                io.remove_directory(root_dir / name)
            raise e

    @classmethod
    def dummy(
        cls,
        name: str,
        task: str,
        image_num: int = 100,
        category_num: int = 10,
        unlabeled_image_num: int = 0,
        root_dir: str = None,
    ) -> "Dataset":
        """
        Create Dummy Dataset (for debugging).

        Args:
            name (str): Dataset name
            task (str): Dataset task
            image_num (int, optional): Number of images. Defaults to 100.
            category_num (int, optional): Number of categories. Defaults to 10.
            unlabeld_image_num (int, optional): Number of unlabeled images. Defaults to 0.
            root_dir (str, optional): Dataset root directory. Defaults to None.

        Raises:
            FileExistsError: if dataset name already exists

        Examples:
            >>> ds = Dataset.dummy("my_dataset", "CLASSIFICATION", image_num=100, category_num=10)
            >>> len(ds.get_images())
            100
            >>> len(ds.get_categories())
            10
        """
        ds = Dataset.new(name=name, task=task, root_dir=root_dir)

        try:
            for category_id in range(1, category_num + 1):

                if task == TaskType.CLASSIFICATION:
                    category = Category.classification(
                        category_id=category_id,
                        name=f"category_{category_id}",
                        supercategory="object",
                    )
                elif task == TaskType.OBJECT_DETECTION:
                    category = Category.object_detection(
                        category_id=category_id,
                        name=f"category_{category_id}",
                        supercategory="object",
                    )
                elif task == TaskType.INSTANCE_SEGMENTATION:
                    category = Category.instance_segmentation(
                        category_id=category_id,
                        name=f"category_{category_id}",
                        supercategory="object",
                    )
                elif task == TaskType.TEXT_RECOGNITION:
                    category = Category.text_recognition(
                        category_id=category_id,
                        name=chr(64 + category_id),
                        supercategory="object",
                    )

                ds.add_categories([category])

            annotation_id = 1
            for image_id in range(1, image_num + 1):
                file_name = f"image_{image_id}.jpg"
                ds.add_images([Image(image_id=image_id, file_name=file_name, width=100, height=100)])
                PIL.Image.new("RGB", (100, 100)).save(ds.raw_image_dir / file_name)

                if task == TaskType.CLASSIFICATION:
                    annotations = [
                        Annotation.classification(
                            annotation_id=annotation_id,
                            image_id=image_id,
                            category_id=random.randint(1, category_num),
                        )
                    ]
                elif task == TaskType.OBJECT_DETECTION:
                    annotations = [
                        Annotation.object_detection(
                            annotation_id=annotation_id + i,
                            image_id=image_id,
                            category_id=random.randint(1, category_num),
                            bbox=[
                                random.randint(0, 100),
                                random.randint(0, 100),
                                random.randint(0, 100),
                                random.randint(0, 100),
                            ],
                        )
                        for i in range(random.randint(1, 5))
                    ]
                elif task == TaskType.INSTANCE_SEGMENTATION:
                    annotations = [
                        Annotation.instance_segmentation(
                            annotation_id=annotation_id + i,
                            image_id=image_id,
                            category_id=random.randint(1, category_num),
                            bbox=[
                                random.randint(0, 100),
                                random.randint(0, 100),
                                random.randint(0, 100),
                                random.randint(0, 100),
                            ],
                            segmentation=[[random.randint(0, 100) for _ in range(10)]],
                        )
                        for i in range(random.randint(1, 5))
                    ]
                elif task == TaskType.TEXT_RECOGNITION:
                    annotations = [
                        Annotation.text_recognition(
                            annotation_id=annotation_id,
                            image_id=image_id,
                            caption=chr(64 + random.randint(1, category_num)),
                        )
                    ]
                ds.add_annotations(annotations)
                annotation_id += len(annotations)

            if unlabeled_image_num > 0:
                for image_id in range(image_num + 1, image_num + unlabeled_image_num + 1):
                    file_name = f"image_{image_id}.jpg"
                    ds.add_images(
                        [Image(image_id=image_id, file_name=file_name, width=100, height=100)]
                    )
                    PIL.Image.new("RGB", (100, 100)).save(ds.raw_image_dir / file_name)

        except Exception as e:
            ds.delete()
            raise e

        ds.create_index()
        return ds

    @classmethod
    def load(cls, name: str, root_dir: str = None) -> "Dataset":
        """
        Load Dataset.
        This method loads an existing dataset.

        Args:
            name (str): Dataset name that Waffle Created
            root_dir (str, optional): Dataset root directory. Defaults to None.

        Raises:
            FileNotFoundError: if source dataset does not exist.

        Examples:
            >>> ds = Dataset.load("my_dataset")
            >>> ds.name
            'my_dataset'  # dataset name

        Returns:
            Dataset: Dataset Class
        """
        root_dir = Dataset.parse_root_dir(root_dir)
        dataset_info_file = root_dir / name / Dataset.DATASET_INFO_FILE_NAME
        if not dataset_info_file.exists():
            raise FileNotFoundError(f"{dataset_info_file} has not been created.")
        dataset_info = DatasetInfo.load(dataset_info_file)

        ds = cls(**dataset_info.to_dict(), root_dir=root_dir)
        ds.create_index()
        return ds

    @classmethod
    def merge(
        cls,
        name: str,
        root_dir: str,
        src_names: list[str],
        src_root_dirs: Union[str, list[str]],
        task: str,
    ) -> "Dataset":
        """
        Merge Datasets.
        This method merges multiple datasets into one dataset.

        Args:
            name (str): New Dataset name
            root_dir (str): New Dataset root directory
            src_names (list[str]): Source Dataset names
            src_root_dirs (Union[str, list[str]]): Source Dataset root directories
            task (str): Dataset task

        Returns:
            Dataset: Dataset Class

        """
        if isinstance(src_root_dirs, str):
            src_root_dirs = [src_root_dirs] * len(src_names)
        if len(src_names) != len(src_root_dirs):
            raise ValueError("Length of src_names and src_root_dirs should be same.")
        if isinstance(task, str):
            task = task.upper()
        if task not in [k for k in TaskType]:
            raise ValueError(f"task should be one of {[k for k in TaskType]}")

        merged_ds = Dataset.new(
            name=name,
            root_dir=root_dir,
            task=task,
        )

        categoryname2id = {}
        filename2id = {}
        new_annotation_id = 1

        try:
            for src_name, src_root_dir in zip(src_names, src_root_dirs):
                src_ds = Dataset.load(src_name, src_root_dir)
                src_categories = src_ds.get_categories()

                if src_ds.task != task:
                    raise ValueError(f"Task of {src_ds.name} is {src_ds.task}. It should be {task}.")

                # merge - raw images
                io.copy_files_to_directory(
                    src_ds.raw_image_dir, merged_ds.raw_image_dir, create_directory=True
                )

                # merge - categories
                for category in src_ds.get_categories():
                    if category.name not in categoryname2id:
                        new_category_id = len(categoryname2id) + 1
                        categoryname2id[category.name] = new_category_id

                        new_category = copy.deepcopy(category)
                        new_category.category_id = new_category_id
                        merged_ds.add_categories([new_category])

                for image_id, annotations in src_ds.image_to_annotations.items():
                    image = src_ds.get_images([image_id])[0]

                    # merge - images
                    is_new_image = False
                    if image.file_name not in filename2id:
                        is_new_image = True

                        new_image_id = len(filename2id) + 1
                        filename2id[image.file_name] = new_image_id

                        new_image = copy.deepcopy(image)
                        new_image.image_id = new_image_id
                        merged_ds.add_images([new_image])

                    new_image_id = filename2id[image.file_name]

                    # merge - annotations
                    for annotation in annotations:
                        new_annotation = copy.deepcopy(annotation)
                        new_annotation.category_id = categoryname2id[
                            src_categories[annotation.category_id - 1].name
                        ]

                        # check if new annotation
                        is_new_annotation = True
                        if not is_new_image:
                            for merged_ann in merged_ds.get_annotations(new_image_id):
                                if new_annotation == merged_ann:
                                    is_new_annotation = False
                                    break

                        # merge
                        if is_new_annotation:
                            new_annotation.image_id = filename2id[image.file_name]
                            new_annotation.annotation_id = new_annotation_id
                            new_annotation_id += 1
                            merged_ds.add_annotations([new_annotation])

        except Exception as e:
            if merged_ds.dataset_dir.exists():
                io.remove_directory(merged_ds.dataset_dir)
            raise e

        ds = Dataset.load(name, root_dir)
        ds.create_index()
        return ds

    @classmethod
    def from_coco(
        cls,
        name: str,
        task: str,
        coco_file: Union[str, list[str]],
        coco_root_dir: Union[str, list[str]],
        root_dir: str = None,
    ) -> "Dataset":
        """
        Import Dataset from coco format.
        This method imports coco format dataset.

        Args:
            name (str): Dataset name.
            task (str): Dataset task.
            coco_file (Union[str, list[str]]): Coco json file path. If given list, it will be regarded as [train, val, test] json file.
            coco_root_dir (Union[str, list[str]]): Coco image root directory. If given list, it will be regarded as [train, val, test] coco root file.
            root_dir (str, optional): Dataset root directory. Defaults to None.

        Raises:
            FileExistsError: if new dataset name already exist.

        Examples:
            # Import one coco json file.
            >>> ds = Dataset.from_coco("my_dataset", "object_detection", "path/to/coco.json", "path/to/coco_root")
            >>> ds.get_images()
            {<Image: 1>, <Image: 2>, <Image: 3>, <Image: 4>, <Image: 5>}
            >>> ds.get_annotations()
            {<Annotation: 1>, <Annotation: 2>, <Annotation: 3>, <Annotation: 4>, <Annotation: 5>}
            >>> ds.get_categories()
            {<Category: 1>, <Category: 2>, <Category: 3>, <Category: 4>, <Category: 5>}
            >>> ds.get_category_names()
            ['person', 'bicycle', 'car', 'motorcycle', 'airplane']

            # Import multiple coco json files.
            # You can give coco_file as list.
            # Given coco files are regarded as [train, [val, [test]]] json files.
            >>> ds = Dataset.from_coco("my_dataset", "object_detection", ["coco_train.json", "coco_val.json"], ["coco_train_root", "coco_val_root"])

        Returns:
            Dataset: Dataset Class
        """
        ds = Dataset.new(name=name, task=task, root_dir=root_dir)

        try:
            if isinstance(coco_file, list) and isinstance(coco_root_dir, list):
                if len(coco_file) != len(coco_root_dir):
                    raise ValueError("coco_file and coco_root_dir should have same length.")
            if not isinstance(coco_file, list) and isinstance(coco_root_dir, list):
                raise ValueError(
                    "ambiguous input. The number of coco_file should be same or greater than coco_root_dir."
                )
            if not isinstance(coco_file, list):
                coco_file = [coco_file]
            if not isinstance(coco_root_dir, list):
                coco_root_dir = [coco_root_dir] * len(coco_file)

            coco_files = coco_file
            coco_root_dirs = coco_root_dir

            import_coco(ds, coco_files, coco_root_dirs)

            if len(coco_files) == 2:
                logger.info("copying val set to test set")
                io.copy_file(ds.val_set_file, ds.test_set_file, create_directory=True)

            # TODO: add unlabeled set
            io.save_json([], ds.unlabeled_set_file, create_directory=True)

        except Exception as e:
            ds.delete()
            raise e

        ds.create_index()
        return ds

    @classmethod
    def from_autocare_dlt(
        cls,
        name: str,
        task: str,
        coco_file: Union[str, list[str]],
        coco_root_dir: Union[str, list[str]],
        root_dir: str = None,
    ) -> "Dataset":
        """
        Import dataset from autocare dlt format.
        This method is used for importing dataset from autocare dlt format.

        Args:
            name (str): name of dataset.
            task (str): task of dataset.
            coco_file (Union[str, list[str]]): coco annotation file path.
            coco_root_dir (Union[str, list[str]]): root directory of coco dataset.
            root_dir (str, optional): root directory of dataset. Defaults to None.

        Raises:
            FileExistsError: if new dataset name already exist.

        Examples:
            # Import one coco json file.
            >>> ds = Dataset.from_coco("my_dataset", "object_detection", "path/to/coco.json", "path/to/coco_root")
            >>> ds.get_images()
            {<Image: 1>, <Image: 2>, <Image: 3>, <Image: 4>, <Image: 5>}
            >>> ds.get_annotations()
            {<Annotation: 1>, <Annotation: 2>, <Annotation: 3>, <Annotation: 4>, <Annotation: 5>}
            >>> ds.get_categories()
            {<Category: 1>, <Category: 2>, <Category: 3>, <Category: 4>, <Category: 5>}
            >>> ds.get_category_names()
            ['person', 'bicycle', 'car', 'motorcycle', 'airplane']

        Returns:
            Dataset: Dataset Class.
        """
        ds = Dataset.new(name=name, task=task, root_dir=root_dir)

        try:
            if isinstance(coco_file, list) and isinstance(coco_root_dir, list):
                if len(coco_file) != len(coco_root_dir):
                    raise ValueError("coco_file and coco_root_dir should have same length.")
            if not isinstance(coco_file, list) and isinstance(coco_root_dir, list):
                raise ValueError(
                    "ambiguous input. The number of coco_file should be same or greater than coco_root_dir."
                )
            if not isinstance(coco_file, list):
                coco_file = [coco_file]
            if not isinstance(coco_root_dir, list):
                coco_root_dir = [coco_root_dir] * len(coco_file)

            coco_files = coco_file
            coco_root_dirs = coco_root_dir

            import_autocare_dlt(ds, coco_files, coco_root_dirs)

            if len(coco_files) == 2:
                logging.info("copying val set to test set")
                io.copy_file(ds.val_set_file, ds.test_set_file, create_directory=True)

            # TODO: add unlabeled set
            io.save_json([], ds.unlabeled_set_file, create_directory=True)

        except Exception as e:
            ds.delete()
            raise e

        ds.create_index()
        return ds

    @classmethod
    def from_yolo(
        cls,
        name: str,
        task: str,
        yolo_root_dir: str,
        yaml_path: str = None,
        root_dir: str = None,
    ) -> "Dataset":
        """
        Import Dataset from yolo format.
        This method imports dataset from yolo(ultralytics) yaml file.

        Args:
            name (str): Dataset name.
            task (str): Dataset task.
            yolo_root_dir (str): Yolo dataset root directory.
            yaml_path (str): Yolo yaml file path. when task is classification, yaml_path is not required.
            root_dir (str, optional): Dataset root directory. Defaults to None.

        Example:
            >>> ds = Dataset.from_yolo("yolo", "classification", "path/to/yolo.yaml")

        Returns:
            Dataset: Imported dataset.
        """

        ds = Dataset.new(name=name, task=task, root_dir=root_dir)

        try:
            import_yolo(ds, yolo_root_dir, yaml_path)

        except Exception as e:
            ds.delete()
            raise e

        ds.create_index()
        return ds

    @classmethod
    def from_transformers(
        cls,
        name: str,
        task: str,
        dataset_dir: str,
        root_dir=None,
    ) -> "Dataset":
        """
        Import Dataset from transformers datasets.
        This method imports transformers dataset from directory.

        Args:
            name (str): Dataset name.
            dataset_dir (str): Transformers dataset directory.
            task (str): Task name.
            root_dir (str, optional): Dataset root directory. Defaults to None.

        Raises:
            FileExistsError: if dataset name already exists
            ValueError: if dataset is not Dataset or DatasetDict

        Examples:
            >>> ds = Dataset.from_transformers("transformers", "object_detection", "path/to/transformers/dataset")

        Returns:
            Dataset: Dataset Class
        """
        ds = Dataset.new(name=name, task=task, root_dir=root_dir)

        try:
            import_transformers(ds, dataset_dir)

            # TODO: add unlabeled set
            io.save_json([], ds.unlabeled_set_file, create_directory=True)

        except Exception as e:
            ds.delete()
            raise e

        ds.create_index()
        return ds

    @classmethod
    def sample(cls, name: str, task: str, root_dir: str = None) -> "Dataset":
        """
        Import sample Dataset.

        Args:
            name (str): Dataset name.
            task (str): Task name.
            root_dir (str, optional): Dataset root directory. Defaults to None.

        Returns:
            Dataset: Dataset Class
        """

        temp_dir = Path(mkdtemp())
        try:
            if task in [
                TaskType.CLASSIFICATION,
                TaskType.OBJECT_DETECTION,
                TaskType.INSTANCE_SEGMENTATION,
            ]:
                url = "https://raw.githubusercontent.com/snuailab/assets/main/waffle/sample_dataset/mnist.zip"
            elif task == TaskType.TEXT_RECOGNITION:
                url = "https://raw.githubusercontent.com/snuailab/assets/main/waffle/sample_dataset/ocr_sample.zip"
            else:
                raise NotImplementedError(f"not supported task: {task}")

            network.get_file_from_url(url, temp_dir / "mnist.zip")
            io.unzip(temp_dir / "mnist.zip", temp_dir)

            ds = Dataset.from_coco(
                name=name,
                root_dir=root_dir,
                task=task,
                coco_file=str(temp_dir / "coco.json"),
                coco_root_dir=str(temp_dir / "images"),
            )
        except Exception as e:
            raise e
        finally:
            shutil.rmtree(temp_dir)

        ds.create_index()
        return ds

    @classmethod
    def get_dataset_list(cls, root_dir: str = None) -> list[str]:
        """
        Get dataset name list in root_dir.

        Args:
            root_dir (str, optional): dataset root directory. Defaults to None.

        Returns:
            list[str]: dataset name list.
        """
        root_dir = Dataset.parse_root_dir(root_dir)

        if not root_dir.exists():
            return []

        dataset_name_list = []
        for dataset_dir in root_dir.iterdir():
            if dataset_dir.is_dir():
                dataset_info_file = dataset_dir / Dataset.DATASET_INFO_FILE_NAME
                if dataset_info_file.exists():
                    dataset_name_list.append(dataset_dir.name)
        return dataset_name_list

    def initialize(self):
        """Initialize Dataset.
        It creates necessary directories under {dataset_root_dir}/{dataset_name}.
        """

        if self.initialized():
            raise FileExistsError(f"{self.name} is already initialized.")

        io.make_directory(self.raw_image_dir)
        io.make_directory(self.image_dir)
        io.make_directory(self.annotation_dir)
        io.make_directory(self.category_dir)

    def initialized(self) -> bool:
        """Check if Dataset has been initialized or not.

        Returns:
            bool:
                initialized -> True
                not initialized -> False
        """
        return self.dataset_info_file.exists()

    def save_dataset_info(self):
        """Save DatasetInfo."""
        DatasetInfo(
            name=self.name,
            task=self.task,
            categories=list(map(lambda x: x.to_dict(), self.categories)),
            created=self.created,
        ).save_yaml(self.dataset_info_file)

    def trainable(self) -> bool:
        """Check if Dataset is trainable or not.

        Returns:
            bool:
                trainable -> True
                not trainable -> False
        """
        num_images_per_category: dict[int, int] = self.get_num_images_per_category()
        for category_id, image_num in num_images_per_category.items():
            if image_num < Dataset.MINIMUM_TRAINABLE_IMAGE_NUM_PER_CATEGORY:
                return False
        return True

    def check_trainable(self):
        """
        Check if Dataset is trainable or not.

        Raises:
            ValueError: if dataset has not enough annotations.
        """
        if not self.trainable():
            raise ValueError(
                "Dataset is not trainable\n"
                + f"Please check if the MINIMUM_TRAINABLE_IMAGE_NUM_PER_CATEGORY={Dataset.MINIMUM_TRAINABLE_IMAGE_NUM_PER_CATEGORY} is satisfied\n"
                + "Your dataset is consisted of\n"
                + "\n".join(
                    [
                        f"  - {category.name}: {self.get_num_images_per_category()[category.category_id]} images"
                        for category in self.get_categories()
                    ]
                )
            )

    def get_dataset_info(self) -> DatasetInfo:
        """Get DatasetInfo.

        Returns:
            DatasetInfo: DatasetInfo
        """
        dataset_info = DatasetInfo.load(self.dataset_info_file)
        if not hasattr(dataset_info, "categories"):
            dataset_info.categories = self.get_categories()
            self.save_dataset_info()
        return dataset_info

    # get
    def get_images(self, image_ids: list[int] = None, labeled: bool = True) -> list[Image]:
        """Get "Image"s.

        Args:
            image_ids (list[int], optional): id list. None for all "Image"s. Defaults to None.
            labeled (bool, optional): get labeled images. False for unlabeled images. Defaults to True.

        Returns:
            list[Image]: "Image" list
        """
        image_files = (
            list(map(lambda x: self.image_dir / (str(x) + ".json"), image_ids))
            if image_ids
            else list(self.image_dir.glob("*.json"))
        )
        labeled_images = []
        unlabeled_images = []
        for image_file in image_files:
            if self.get_annotations(image_file.stem):
                labeled_images.append(Image.from_json(image_file))
            else:
                unlabeled_images.append(Image.from_json(image_file))

        if labeled:
            logger.info(f"Found {len(labeled_images)} labeled images")
            return labeled_images
        else:
            logger.info(f"Found {len(unlabeled_images)} unlabeled images")
            return unlabeled_images

    def get_categories(self, category_ids: list[int] = None) -> list[Category]:
        """Get "Category"s.

        Args:
            category_ids (list[int], optional): id list. None for all "Category"s. Defaults to None.

        Returns:
            list[Category]: "Category" list
        """
        return sorted(
            [
                Category.from_json(f, self.task)
                for f in (
                    [self.category_dir / f"{category_id}.json" for category_id in category_ids]
                    if category_ids
                    else self.category_dir.glob("*.json")
                )
            ],
            key=lambda x: x.category_id,
        )

    def get_annotations(self, image_id: int = None) -> list[Annotation]:
        """Get "Annotation"s.

        Args:
            image_id (int, optional): image id. None for all "Annotation"s. Defaults to None.

        Returns:
            list[Annotation]: "Annotation" list
        """
        if image_id:
            return [
                Annotation.from_json(f, self.task)
                for f in self.annotation_dir.glob(f"{image_id}/*.json")
            ]
        else:
            return [Annotation.from_json(f, self.task) for f in self.annotation_dir.glob("*/*.json")]

    def get_predictions(self, image_id: int = None) -> list[Annotation]:
        """Get "Prediction"s.

        Args:
            image_id (int, optional): image id. None for all "Prediction"s. Defaults to None.

        Returns:
            list[Annotation]: "Prediction" list
        """
        if image_id:
            return [
                Annotation.from_json(f, self.task)
                for f in self.prediction_dir.glob(f"{image_id}/*.json")
            ]
        else:
            return [Annotation.from_json(f, self.task) for f in self.prediction_dir.glob("*/*.json")]

    def get_num_images_per_category(self) -> dict[int, int]:
        self.num_images_per_category = {
            category_id: len(images) for category_id, images in self.category_to_images.items()
        }
        return self.num_images_per_category

    def get_num_annotations_per_category(self) -> dict[int, int]:
        num_annotations_per_category = {
            category_id: len(annotations)
            for category_id, annotations in self.category_to_annotations.items()
        }
        return num_annotations_per_category

    def create_index(self):
        """Create index for faster search."""
        self._image_dict = OrderedDict()
        self._unlabeled_image_dict = OrderedDict()
        self._annotation_dict = OrderedDict()
        self._prediction_dict = OrderedDict()
        self._category_dict = OrderedDict()
        self._image_to_annotations = OrderedDict()
        self._image_to_predictions = OrderedDict()
        self._annotation_to_image = OrderedDict()
        self._prediction_to_image = OrderedDict()
        self._category_to_images = OrderedDict()
        self._category_to_unique_images = OrderedDict()
        self._category_name_to_category = OrderedDict()
        self._category_to_annotations = OrderedDict()
        self._category_to_predictions = OrderedDict()

        start = time.time()
        logger.info("Creating index for faster search")

        for image in self.get_images():
            self._image_dict[image.image_id] = image  # image_id: image
            self._image_to_annotations[image.image_id] = []
            self._image_to_predictions[image.image_id] = []

        for annotation in self.get_annotations():
            self._annotation_dict[annotation.annotation_id] = annotation  # annotation_id: annotation
            self._image_to_annotations[annotation.image_id].append(
                annotation
            )  # image_id: [annotations]
            self._annotation_to_image[annotation.annotation_id] = self._image_dict[
                annotation.image_id
            ]  # annotation_id: image

        for prediction in self.get_predictions():
            self._prediction_dict[prediction.annotation_id] = prediction  # annotation_id: prediction
            self._image_to_predictions[prediction.image_id].append(
                prediction
            )  # image_id: [predictions]
            self._prediction_to_image[prediction.annotation_id] = self._image_dict[
                prediction.image_id
            ]  # prediction_id: image

        for category in self.get_categories():
            self._category_dict[category.category_id] = category  # category_id: category
            self._category_name_to_category[category.name] = category  # category_name: category
            self._category_to_unique_images[category.category_id] = []  # category_id: image
            self._category_to_images[category.category_id] = set()
            self._category_to_annotations[category.category_id] = []

        for annotation in self._annotation_dict.values():
            if self.task == TaskType.TEXT_RECOGNITION:
                chars = set(annotation.caption)
                for char in chars:
                    category_id = self._category_name_to_category[char].category_id
                    self._category_to_annotations[category_id].append(annotation)
                    self._category_to_images[category_id].add(
                        self._annotation_to_image[annotation.annotation_id]
                    )
            else:
                self._category_to_annotations[annotation.category_id].append(
                    annotation
                )  # category_id: [annotations]
                self._category_to_images[annotation.category_id].add(
                    self._annotation_to_image[annotation.annotation_id]
                )  # category_id: {images}

        for image_id, annotations in self._image_to_annotations.items():
            if self.task == TaskType.TEXT_RECOGNITION:
                most_common_category = Counter(
                    sum([list(annotation.caption) for annotation in annotations], [])
                ).most_common(1)[0][0]
                most_common_category_id = self._category_name_to_category[
                    most_common_category
                ].category_id
                self._category_to_unique_images[most_common_category_id].append(
                    self._image_dict[image_id]
                )
            else:
                most_common_category_id = Counter(
                    [annotation.category_id for annotation in annotations]
                ).most_common(1)[0][0]
                self._category_to_unique_images[most_common_category_id].append(
                    self._image_dict[image_id]
                )

        for category_id, images in self._category_to_images.items():
            self._category_to_images[category_id] = list(images)

        for prediction in self._prediction_dict.values():
            self._category_to_predictions[prediction.category_id].append(
                prediction
            )  # category_id: [predictions]

        for image in self.get_images(labeled=False):
            self._unlabeled_image_dict[image.image_id] = image  # unlabeled_image_id: image

        logger.info(f"Creating index done {time.time() - start:.2f} seconds")

    # add
    def add_images(self, images: Union[Image, list[Image]]):
        """Add "Image"s to dataset.

        Args:
            images (Union[Image, list[Image]]): list of "Image"s
        """
        if not isinstance(images, list):
            images = [images]

        for item in images:
            item_id = item.image_id
            item_path = self.image_dir / f"{item_id}.json"
            io.save_json(item.to_dict(), item_path)

    def add_categories(self, categories: Union[Category, list[Category]]):
        """Add "Category"s to dataset.

        Args:
            categories (Union[Category, list[Category]]): list of "Category"s
        """
        if not isinstance(categories, list):
            categories = [categories]

        category_names_list = [category.name for category in categories]
        category_names = set(category_names_list)
        if (
            len(category_names) != len(category_names_list)
            or set(self.get_category_names()) & category_names
        ):
            raise ValueError("Category names should be unique")

        for item in categories:
            item_id = item.category_id
            item_path = self.category_dir / f"{item_id}.json"
            io.save_json(item.to_dict(), item_path)

        self.save_dataset_info()

    def add_annotations(self, annotations: Union[Annotation, list[Annotation]]):
        """Add "Annotation"s to dataset.

        Args:
            annotations (Union[Annotation, list[Annotation]]): list of "Annotation"s
        """
        if not isinstance(annotations, list):
            annotations = [annotations]

        categories = self.get_category_names()
        for item in annotations:
            if self.task == TaskType.TEXT_RECOGNITION:
                for char in item.caption:
                    if char not in categories:
                        raise ValueError(f"Category '{char}' is not in dataset")
            item_path = self.annotation_dir / f"{item.image_id}" / f"{item.annotation_id}.json"
            io.save_json(item.to_dict(), item_path, create_directory=True)

    def add_predictions(self, predictions: Union[Annotation, list[Annotation]]):
        """Add "Annotation"s to dataset.

        Args:
            annotations (Union[Annotation, list[Annotation]]): list of "Annotation"s
        """
        if not isinstance(predictions, list):
            predictions = [predictions]

        categories = self.get_category_names()
        for item in predictions:
            if self.task == TaskType.TEXT_RECOGNITION:
                for char in item.caption:
                    if char not in categories:
                        raise ValueError(f"Category '{char}' is not in dataset")
            item_path = self.prediction_dir / f"{item.image_id}" / f"{item.annotation_id}.json"
            io.save_json(item.to_dict(), item_path, create_directory=True)

    # functions
    def split(
        self,
        train_ratio: float,
        val_ratio: float = 0.0,
        test_ratio: float = 0.0,
        method: Union[str, SplitMethod] = SplitMethod.RANDOM,
        seed: int = 0,
    ):
        """
        Split Dataset to train, validation, test, (unlabeled) sets.

        Args:
            train_ratio (float): train num ratio (0 ~ 1).
            val_ratio (float, optional): val num ratio (0 ~ 1).
            test_ratio (float, optional): test num ratio (0 ~ 1).
            method (Union[str, SplitMethod], optional): split method. Defaults to SplitMethod.RANDOM.
            seed (int, optional): random seed. Defaults to 0.

        Raises:
            ValueError: if train_ratio is not between 0.0 and 1.0.
            ValueError: if train_ratio + val_ratio + test_ratio is not 1.0.

        Examples:
            >>> dataset = Dataset.load("some_dataset")
            >>> dataset.split(train_ratio=0.8, val_ratio=0.1, test_ratio=0.1)
            >>> dataset.get_split_ids()
            [[1, 2, 3, 4, 5, 6, 7, 8], [9], [10], []]  # train, val, test, unlabeled image ids
        """

        self.check_trainable()

        if train_ratio <= 0.0 or train_ratio >= 1.0:
            raise ValueError(
                "train_ratio must be between 0.0 and 1.0\n" f"given train_ratio: {train_ratio}"
            )

        if val_ratio == 0.0 and test_ratio == 0.0:
            val_ratio = 1 - train_ratio

        if train_ratio + val_ratio + test_ratio != 1.0:
            raise ValueError(
                "train_ratio + val_ratio + test_ratio must be 1.0\n"
                f"given train_ratio: {train_ratio}, val_ratio: {val_ratio}, test_ratio: {test_ratio}"
            )

        if method == SplitMethod.RANDOM:
            random.seed(seed)

            train_ids = []
            val_ids = []
            test_ids = []

            for category_id, images in self.category_to_unique_images.items():
                image_num = len(images)
                image_ids = list(map(lambda x: x.image_id, images))
                random.shuffle(image_ids)

                # flatten images to one list [cat]
                train_num = max(int(image_num * train_ratio), 1)
                val_num = max(int(image_num * val_ratio), 1)

                if test_ratio == 0.0:
                    train_ids.extend(image_ids[:train_num])
                    val_ids.extend(image_ids[train_num:])
                else:
                    train_ids.extend(image_ids[:train_num])
                    val_ids.extend(image_ids[train_num : train_num + val_num])
                    test_ids.extend(image_ids[train_num + val_num :])

        else:
            raise ValueError(f"Unknown split method: {method}")

        logger.info(
            f"train num: {len(train_ids)}  val num: {len(val_ids)}  test num: {len(test_ids)}"
        )

        io.save_json(
            train_ids,
            self.train_set_file,
            create_directory=True,
        )
        io.save_json(
            val_ids,
            self.val_set_file,
            create_directory=True,
        )
        io.save_json(
            test_ids,
            self.test_set_file,
            create_directory=True,
        )

        unlabeled_ids = [img.image_id for img in self.get_images(labeled=False)]

        io.save_json(
            unlabeled_ids,
            self.unlabeled_set_file,
            create_directory=True,
        )

    # export
    def get_split_ids(self) -> list[list[int]]:
        """
        Get split ids

        Returns:
            list[list[int]]: split ids
        """
        if not self.train_set_file.exists():
            raise FileNotFoundError("There is no set files. Please run ds.split() first")

        train_ids: list[int] = (
            io.load_json(self.train_set_file) if self.train_set_file.exists() else []
        )
        val_ids: list[int] = io.load_json(self.val_set_file) if self.val_set_file.exists() else []
        test_ids: list[int] = io.load_json(self.test_set_file) if self.test_set_file.exists() else []
        unlabeled_ids: list[int] = (
            io.load_json(self.unlabeled_set_file) if self.unlabeled_set_file.exists() else []
        )

        return [train_ids, val_ids, test_ids, unlabeled_ids]

    def export(self, data_type: Union[str, DataType]) -> str:
        """
        Export Dataset to Specific data formats

        Args:
            data_type (Union[str, DataType]): export data type. one of ["YOLO", "COCO"].

        Raises:
            ValueError: if data_type is not one of DataType.

        Examples:
            >>> dataset = Dataset.load("some_dataset")
            >>> dataset.export(data_type="YOLO")
            path/to/dataset_dir/exports/yolo

            # You can train with exported dataset
            >>> hub.train("path/to/dataset_dir/exports/yolo", ...)

        Returns:
            str: exported dataset directory
        """

        self.check_trainable()

        export_dir: Path = self.export_dir / EXPORT_MAP[data_type.upper()]
        if data_type in [DataType.YOLO, DataType.ULTRALYTICS]:
            export_function = export_yolo
        elif data_type in [DataType.COCO]:
            export_function = export_coco
        elif data_type in [DataType.AUTOCARE_DLT]:
            export_function = export_autocare_dlt
        elif data_type in [DataType.TRANSFORMERS]:
            export_function = export_transformers

        else:
            raise ValueError(f"Invalid data_type: {data_type}")

        try:
            if export_dir.exists():
                io.remove_directory(export_dir)
                warnings.warn(f"{export_dir} already exists. Removing exist export and override.")

            export_dir = export_function(self, export_dir)

            return export_dir

        except Exception as e:
            if export_dir.exists():
                io.remove_directory(export_dir)
            raise e

    def delete(self):
        """Delete Dataset"""
        io.remove_directory(self.dataset_dir)
        del self

    def draw_annotations(self, image_ids=None):
        """
        Draw annotations on images
        Save drawn images to draw_dir

        Args:
            image_ids (list[int], optional): image ids to draw. Defaults to None.

        """
        if not self.draw_dir.exists():
            self.draw_dir.mkdir(parents=True)

        images = self.get_images(image_ids)
        names = self.get_category_names()

        for image in tqdm.tqdm(images):
            np_image = load_image(self.raw_image_dir / image.file_name)
            annotations = self.get_annotations(image.image_id)
            drawn_image = draw_results(np_image, annotations, names)
            save_image(self.draw_dir / image.file_name, drawn_image)



from waffle_hub import TaskType
from waffle_hub.schema.configs import TrainConfig

# Common
MODEL_TYPES = {
    TaskType.OBJECT_DETECTION: {"YOLOv5": list("sml")},
    TaskType.CLASSIFICATION: {"Classifier": list("sml")},
    TaskType.TEXT_RECOGNITION: {
        "TextRecognition": list("sml"),
        "LicencePlateRecognition": list("sml"),
    },
}

# Backend Specifics
DATA_TYPE_MAP = {
    TaskType.OBJECT_DETECTION: "COCODetectionDataset",
    TaskType.CLASSIFICATION: "COCOClassificationDataset",
    TaskType.TEXT_RECOGNITION: "COCOTextRecognitionDataset",
}

WEIGHT_PATH = {
    TaskType.OBJECT_DETECTION: {
        "YOLOv5": {
            "s": "temp/autocare_dlt/detectors/small/model.pth",
            "m": "temp/autocare_dlt/detectors/medium/model.pth",
            "l": "temp/autocare_dlt/detectors/large/model.pth",
        }
    },
    TaskType.CLASSIFICATION: {
        "Classifier": {
            "s": "temp/autocare_dlt/classifiers/small/model.pth",
            "m": "temp/autocare_dlt/classifiers/medium/model.pth",
            "l": "temp/autocare_dlt/classifiers/large/model.pth",
        }
    },
    TaskType.TEXT_RECOGNITION: {
        "TextRecognition": {
            "s": "temp/autocare_dlt/text_recognizers/small/model.pth",
            "m": "temp/autocare_dlt/text_recognizers/small/model.pth",
            "l": "temp/autocare_dlt/text_recognizers/small/model.pth",
        },
        "LicencePlateRecognition": {
            "s": "temp/autocare_dlt/text_recognizers/small/model.pth",
            "m": "temp/autocare_dlt/text_recognizers/small/model.pth",
            "l": "temp/autocare_dlt/text_recognizers/small/model.pth",
        },
    },
}

DEFAULT_PARAMS = {
    TaskType.OBJECT_DETECTION: {
        "YOLOv5": {
            "s": TrainConfig(
                epochs=50,
                image_size=[640, 640],
                learning_rate=0.01,
                letter_box=True,
                batch_size=32,
            ),
            "m": TrainConfig(
                epochs=50,
                image_size=[640, 640],
                learning_rate=0.01,
                letter_box=True,
                batch_size=16,
            ),
            "l": TrainConfig(
                epochs=50,
                image_size=[640, 640],
                learning_rate=0.01,
                letter_box=True,
                batch_size=8,
            ),
        }
    },
    TaskType.CLASSIFICATION: {
        "Classifier": {
            "s": TrainConfig(
                epochs=50,
                image_size=[224, 224],
                learning_rate=0.01,
                letter_box=False,
                batch_size=256,
            ),
            "m": TrainConfig(
                epochs=50,
                image_size=[224, 224],
                learning_rate=0.01,
                letter_box=False,
                batch_size=128,
            ),
            "l": TrainConfig(
                epochs=50,
                image_size=[224, 224],
                learning_rate=0.01,
                letter_box=False,
                batch_size=64,
            ),
        }
    },
    TaskType.TEXT_RECOGNITION: {
        "TextRecognition": {
            "s": TrainConfig(
                epochs=50,
                image_size=[224, 224],
                learning_rate=0.01,
                letter_box=False,
                batch_size=256,
            ),
            "m": TrainConfig(
                epochs=50,
                image_size=[224, 224],
                learning_rate=0.01,
                letter_box=False,
                batch_size=128,
            ),
            "l": TrainConfig(
                epochs=50,
                image_size=[224, 224],
                learning_rate=0.01,
                letter_box=False,
                batch_size=64,
            ),
        },
        "LicencePlateRecognition": {
            "s": TrainConfig(
                epochs=50,
                image_size=[240, 80],
                learning_rate=0.0002,
                letter_box=False,
                batch_size=256,
            ),
            "m": TrainConfig(
                epochs=50,
                image_size=[240, 80],
                learning_rate=0.0002,
                letter_box=False,
                batch_size=128,
            ),
            "l": TrainConfig(
                epochs=50,
                image_size=[240, 80],
                learning_rate=0.0002,
                letter_box=False,
                batch_size=64,
            ),
        },
    },
}



import logging
import re
import time
from abc import ABC, abstractmethod
from pathlib import Path
from threading import Thread
from typing import Union

__all__ = ["MetricLogger"]

logger = logging.getLogger(__name__)


metric_logger_classes = []


class _BaseMetricLogger(ABC):
    def __init__(self, name, log_dir, **kwargs):
        self.name = str(name)
        self.log_dir = Path(log_dir)

    @abstractmethod
    def log_metric(self, tag, value, step):
        raise NotImplementedError

    @abstractmethod
    def open(self):
        raise NotImplementedError

    @abstractmethod
    def close(self):
        raise NotImplementedError


try:
    from torch.utils.tensorboard import SummaryWriter

    class _TensorboardLogger(_BaseMetricLogger):
        def __init__(self, name, log_dir, **kwargs):
            super().__init__(name, log_dir, **kwargs)

            self.writer = None

        def log_metric(self, tag, value, step):
            if self.writer is None:
                raise RuntimeError("Tensorboard logger is not opened.")
            self.writer.add_scalar(tag, value, step)

        def open(self):
            self.writer = SummaryWriter(self.log_dir)
            logging.info(f"Tensorboard logs will be saved to {self.log_dir}")

        def close(self):
            self.writer.close()
            self.writer = None

    metric_logger_classes.append(_TensorboardLogger)

except ImportError:
    logger.warning(
        "Tensorboard is not installed. To use Tensorboard logger, please install it with `pip install tensorboard`"
    )

try:
    import mlflow

    class _MLFlowLogger(_BaseMetricLogger):
        def __init__(self, name, log_dir, **kwargs):
            super().__init__(name, log_dir, **kwargs)

            self.kwargs = kwargs

        def log_metric(self, tag, value, step):
            # remove invalid characters in tag using regex (only alphabet, -, _, / are allowed)
            tag = re.sub(r"[^a-zA-Z0-9-_\/]", "_", tag)
            with mlflow.start_run(run_name=self.kwargs.get("run_name", "0")):
                mlflow.log_metric(tag, value, step)

        def open(self):
            exp = mlflow.set_experiment(self.name)
            # TODO: only support one model for one experiment
            for run in mlflow.search_runs(experiment_ids=exp.experiment_id).run_id:
                mlflow.delete_run(mlflow.get_run(run).info.run_id)
            logging.info(f"MLFlow logs will be saved to {mlflow.get_tracking_uri()}")

        def close(self):
            mlflow.end_run()

    metric_logger_classes.append(_MLFlowLogger)

except ImportError:
    logger.warning(
        "MLFlow is not installed. To use MLFlow logger, please install it with `pip install mlflow`"
    )


class MetricLogger:
    STEP_KEYWORDS = ["step", "epoch", "iter", "iteration"]
    MAX_TRY = 10

    def __init__(
        self,
        name: str,
        log_dir: Union[str, Path],
        func: callable,
        interval: float,
        prefix: str = "",
        **kwargs,
    ):
        """A metric logger that logs metrics every interval seconds.

        Args:
            name (str): The name of the metric logger.
            log_dir (Union[str, Path]): The directory to save the logs.
            func (Callable): The function to get the metrics.
                func should return a list of metrics.
                e.g. func() -> [[{"tag": "value}, ...], ...]
            interval (float): The interval to log metrics. (seconds)
            prefix (str, optional): The prefix of the log file. Defaults to "".
            kwargs: The arguments for the metric logger.
        """

        self.name = str(name)
        self.log_dir = log_dir
        self.func = func
        self.interval = float(interval)
        self.prefix = str(prefix)

        self.kwargs = kwargs

        self.loggers = [
            logger(self.name, self.log_dir, **kwargs) for logger in metric_logger_classes
        ]

        self._stop = False
        self._thread = None
        self._initiated = False
        self._last_step = 0

    def log_metric(self, tag, value, step):
        for logger in self.loggers:
            logger.log_metric((self.prefix + "/" if self.prefix else "") + tag, value, step)

    def start(self):
        """Start logging thread."""
        if self._initiated:
            raise RuntimeError("Metric logger has already been initiated.")

        self._initiated = True

        for logger in self.loggers:
            logger.open()

        self._stop = False

        self.thread = Thread(target=self._loop, daemon=True)
        self.thread.start()

    def stop(self):
        """Stop logging thread."""
        self._stop = True

        self.thread.join()
        for logger in self.loggers:
            logger.close()

    def _loop(self):
        """Log metrics every interval seconds."""
        while not self._stop:
            chance = self.MAX_TRY
            while True:
                try:
                    self._log()
                    break
                except Exception as e:
                    chance -= 1
                    if chance == 0:
                        raise e
                    time.sleep(1)
            time.sleep(self.interval)
        self._log()  # final log after stop

    def _log(self):
        """Log metrics."""
        metrics_per_epoch = self.func()
        current_step = len(metrics_per_epoch)
        for step in range(self._last_step, current_step):
            # metrics is a list of dict
            # e.g. [{"tag": "value"}, ...]
            metric_dict = {v["tag"]: v["value"] for v in metrics_per_epoch[step]}

            # pop all keys from metrics if they contains step keywords
            for key in list(metric_dict.keys()):
                for step_keyword in self.STEP_KEYWORDS:
                    if step_keyword == key.lower():
                        metric_dict.pop(key)

            for tag, value in metric_dict.items():
                self.log_metric(tag, value, step)

        self._last_step = current_step



import warnings
from collections import defaultdict
from itertools import groupby
from pathlib import Path
from typing import Union

import cv2
from waffle_utils.file import io, search

from waffle_hub import TaskType
from waffle_hub.schema.fields import Annotation, Category, Image
from waffle_hub.utils.conversion import merge_multi_segment


def _check_valid_file_paths(images: list[Image]) -> bool:
    """Check file paths are valid
    If the file name includes the words "images" or "labels," an error occurs during training

    Args:
        images (list[Image]): List of Image

    Returns:
        bool: True if valid
    """
    for image in images:
        file_path = Path(image.file_name)
        if "images" in file_path.parts:
            raise ValueError(
                f"The file path '{file_path}' is not allowed. Please choose a file path that does not contain the word 'images'"
            )
        if "labels" in file_path.parts:
            raise ValueError(
                f"The file path '{file_path}' is not allowed. Please choose a file path that does not contain the word 'labels'"
            )
    else:
        return True


def _export_yolo_classification(
    self,
    export_dir: Path,
    train_ids: list,
    val_ids: list,
    test_ids: list,
    unlabeled_ids: list,
):
    """Export dataset to YOLO format for classification task

    Args:
        export_dir (Path): Path to export directory
        train_ids (list): List of train ids
        val_ids (list): List of validation ids
        test_ids (list): List of test ids
        unlabeled_ids (list): List of unlabeled ids
    """
    io.make_directory(export_dir)

    for split, image_ids in zip(
        ["train", "val", "test", "unlabeled"],
        [train_ids, val_ids, test_ids, unlabeled_ids],
    ):
        if len(image_ids) == 0:
            continue

        split_dir = export_dir / split
        io.make_directory(split_dir)

        category_names = {c.category_id: c.name for c in self.get_categories()}
        for image in self.get_images(image_ids):
            image_path = self.raw_image_dir / image.file_name

            annotations = self.get_annotations(image.image_id)
            if len(annotations) > 1:
                warnings.warn(f"Multi label does not support yet. Skipping {image_path}.")
                continue
            category_id = annotations[0].category_id

            image_dst_path = split_dir / category_names[category_id] / image.file_name
            io.copy_file(image_path, image_dst_path, create_directory=True)


def _export_yolo_detection(
    self,
    export_dir: Path,
    train_ids: list,
    val_ids: list,
    test_ids: list,
    unlabeled_ids: list,
):
    """Export dataset to YOLO format for detection task

    Args:
        export_dir (Path): Path to export directory
        train_ids (list): List of train ids
        val_ids (list): List of validation ids
        test_ids (list): List of test ids
        unlabeled_ids (list): List of unlabeled ids
    """
    io.make_directory(export_dir)

    for split, image_ids in zip(
        ["train", "val", "test", "unlabeled"],
        [train_ids, val_ids, test_ids, unlabeled_ids],
    ):
        if len(image_ids) == 0:
            continue

        image_dir = export_dir / split / "images"
        label_dir = export_dir / split / "labels"

        io.make_directory(image_dir)
        io.make_directory(label_dir)

        for image in self.get_images(image_ids):
            image_path = self.raw_image_dir / image.file_name
            image_dst_path = image_dir / image.file_name
            label_dst_path = (label_dir / image.file_name).with_suffix(".txt")
            io.copy_file(image_path, image_dst_path, create_directory=True)

            W = image.width
            H = image.height

            annotations = self.get_annotations(image.image_id)
            label_txts = []
            for annotation in annotations:
                x1, y1, w, h = annotation.bbox
                x1, w = x1 / W, w / W
                y1, h = y1 / H, h / H
                cx, cy = x1 + w / 2, y1 + h / 2

                category_id = annotation.category_id - 1

                label_txts.append(f"{category_id} {cx} {cy} {w} {h}")

            io.make_directory(label_dst_path.parent)
            with open(label_dst_path, "w") as f:
                f.write("\n".join(label_txts))


def _export_yolo_segmentation(
    self,
    export_dir: Path,
    train_ids: list,
    val_ids: list,
    test_ids: list,
    unlabeled_ids: list,
):
    io.make_directory(export_dir)

    for split, image_ids in zip(
        ["train", "val", "test", "unlabeled"],
        [train_ids, val_ids, test_ids, unlabeled_ids],
    ):
        if len(image_ids) == 0:
            continue

        image_dir = export_dir / split / "images"
        label_dir = export_dir / split / "labels"

        io.make_directory(image_dir)
        io.make_directory(label_dir)

        for image in self.get_images(image_ids):
            image_path = self.raw_image_dir / image.file_name
            image_dst_path = image_dir / image.file_name
            label_dst_path = (label_dir / image.file_name).with_suffix(".txt")
            io.copy_file(image_path, image_dst_path, create_directory=True)

            W = image.width
            H = image.height

            annotations = self.get_annotations(image.image_id)
            label_txts = []
            for annotation in annotations:
                x1, y1, w, h = annotation.bbox
                x1, w = x1 / W, w / W
                y1, h = y1 / H, h / H

                category_id = annotation.category_id - 1

                segment = merge_multi_segment(annotation.segmentation, (W, H))
                segment[0::2] = [x / W for x in segment[0::2]]
                segment[1::2] = [y / H for y in segment[1::2]]
                segment = " ".join(map(str, segment))

                label_txts.append(f"{category_id} {segment}")

            io.make_directory(label_dst_path.parent)
            with open(label_dst_path, "w") as f:
                f.write("\n".join(label_txts))


def export_yolo(self, export_dir: Union[str, Path]) -> str:
    """Export dataset to YOLO format

    Args:
        export_dir (Union[str, Path]): Path to export directory

    Returns:
        str: Path to export directory
    """
    _check_valid_file_paths(self.get_images())

    export_dir = Path(export_dir)

    train_ids, val_ids, test_ids, _ = self.get_split_ids()

    if self.task == TaskType.CLASSIFICATION:
        _export_yolo_classification(self, export_dir, train_ids, val_ids, test_ids, [])
    elif self.task == TaskType.OBJECT_DETECTION:
        _export_yolo_detection(self, export_dir, train_ids, val_ids, test_ids, [])
    elif self.task == TaskType.INSTANCE_SEGMENTATION:
        _export_yolo_segmentation(self, export_dir, train_ids, val_ids, test_ids, [])
    else:
        raise ValueError(f"Unsupported task type: {self.task}")

    io.save_yaml(
        {
            "path": str(export_dir.absolute()),
            "train": "train",
            "val": "val",
            "test": "test" if test_ids else "val",
            "names": {category.category_id - 1: category.name for category in self.get_categories()},
        },
        export_dir / "data.yaml",
    )

    return str(export_dir)


# import functions
def _get_yolo_image_rel_paths(yolo_root_dir: Path, set_paths: list = ["train", "val", "test"]):
    set_image_rel_path = {}
    for set_path in set_paths:
        set_image_rel_path[set_path] = list(
            map(
                lambda x: x.relative_to(yolo_root_dir),
                search.get_image_files(yolo_root_dir / set_path),
            )
        )
    return set_image_rel_path


def _import_yolo_classification(self, yolo_root_dir: Path, *args):
    # get image relative paths
    set_image_rel_path = _get_yolo_image_rel_paths(yolo_root_dir, ["train", "val", "test"])

    # get categories
    category_group = groupby(set_image_rel_path["train"], lambda image_path: image_path.parts[1])
    category_names = {k for k, _ in category_group}
    category_name2id = {}
    for category_id, category_name in enumerate(category_names, start=1):
        category_name2id[category_name] = category_id
        self.add_categories(
            [
                Category.classification(
                    category_id=category_id,
                    name=category_name,
                )
            ]
        )

    # convert to waffle format
    image_id = 1
    annotation_id = 1
    for set_name, image_rel_paths in set_image_rel_path.items():
        set_image_ids = []
        for image_rel_path in image_rel_paths:
            image_path = yolo_root_dir / image_rel_path
            file_name = f"{image_id}{image_path.suffix}"

            height, width = cv2.imread(str(image_path)).shape[:2]
            self.add_images(
                [
                    Image.new(
                        image_id=image_id,
                        file_name=file_name,
                        width=width,
                        height=height,
                        original_file_name=image_rel_path,
                    )
                ]
            )
            set_image_ids.append(image_id)
            io.copy_file(image_path, self.raw_image_dir / file_name, create_directory=True)

            category_name = image_rel_path.parts[
                1
            ]  # image rel path: {set_name}/{category_name}/{file_name}
            self.add_annotations(
                [
                    Annotation.classification(
                        annotation_id=annotation_id,
                        image_id=image_id,
                        category_id=category_name2id[category_name],
                    )
                ]
            )

            image_id += 1
            annotation_id += 1

        io.save_json(set_image_ids, self.set_dir / f"{set_name}.json", True)

    return True


def _parse_od_label(line: str, width: int, height: int):
    """parse label file for od"""
    category_id, cx, cy, w, h = map(float, line.split())
    x1 = cx - w / 2
    y1 = cy - h / 2
    return {
        "category_id": int(category_id) + 1,
        "bbox": [x1 * width, y1 * height, w * width, h * height],
    }


def _parse_seg_label(line: str, width: int, height: int):
    """parse label file for seg"""
    category_id, *segment = map(float, line.split())
    segment[::2] = [int(x * width) for x in segment[::2]]
    segment[1::2] = [y * height for y in segment[1::2]]
    return {
        "category_id": int(category_id) + 1,
        "segmentation": [segment],
    }


def _import_yolo_images_labels(self, yolo_root_dir: Path, yaml_path: str, task: TaskType):
    """import function for od, seg, keypoint"""

    parse_func = {
        TaskType.OBJECT_DETECTION: _parse_od_label,
        TaskType.INSTANCE_SEGMENTATION: _parse_seg_label,
    }[task]

    info = io.load_yaml(yaml_path)

    # get image relative paths
    set_image_rel_path = _get_yolo_image_rel_paths(
        yolo_root_dir, list(set(map(lambda x: info.get(x, None), ["train", "val", "test"])))
    )

    # get categories
    names = info["names"]
    category_name2id = {}
    if isinstance(names, list):
        names = {category_id: category_name for category_id, category_name in enumerate(names)}
    for category_id, category_name in names.items():
        category_name2id[category_name] = category_id + 1
        self.add_categories(
            [
                Category.new(
                    category_id=category_id + 1,
                    name=category_name,
                    task=task,
                )
            ]
        )

    # convert to waffle format
    image_id = 1
    annotation_id = 1
    for set_name, image_rel_paths in set_image_rel_path.items():
        set_image_ids = []
        for image_rel_path in image_rel_paths:
            image_path = (
                yolo_root_dir / image_rel_path
            )  # image rel path: {set_name}/images/{file_name(.EXT)}
            file_name = f"{image_id}{image_path.suffix}"

            height, width = cv2.imread(str(image_path)).shape[:2]
            self.add_images(
                [
                    Image.new(
                        image_id=image_id,
                        file_name=file_name,
                        width=width,
                        height=height,
                        original_file_name=image_rel_path,
                    )
                ]
            )
            set_image_ids.append(image_id)
            io.copy_file(image_path, self.raw_image_dir / file_name, create_directory=True)

            label_path = image_rel_path.with_suffix(
                ".txt"
            )  # label rel path: {set_name}/labels/{file_name(.txt)}
            label_parts = list(label_path.parts)
            label_parts[1] = "labels"
            label_path = yolo_root_dir / Path(*label_parts)
            with label_path.open("r") as f:
                lines = f.readlines()

            for line in lines:
                annotation = parse_func(line, width, height)
                self.add_annotations(
                    [
                        Annotation.new(
                            annotation_id=annotation_id,
                            image_id=image_id,
                            task=task,
                            **annotation,
                        )
                    ]
                )
                annotation_id += 1
            image_id += 1

        io.save_json(set_image_ids, self.set_dir / f"{set_name}.json", True)


def import_yolo(self, yolo_root_dir: str, yaml_path: str):
    """
    Import YOLO dataset.

    Args:
        yaml_path (str): Path to the yaml file.
    """
    if self.task == TaskType.OBJECT_DETECTION:
        _import = _import_yolo_images_labels
    elif self.task == TaskType.CLASSIFICATION:
        _import = _import_yolo_classification
    elif self.task == TaskType.INSTANCE_SEGMENTATION:
        _import = _import_yolo_images_labels
    else:
        raise ValueError(f"Unsupported task: {self.task}")

    _import(self, Path(yolo_root_dir), yaml_path, self.task)

    # TODO: add unlabeled set
    io.save_json([], self.unlabeled_set_file, create_directory=True)



from dataclasses import dataclass

import numpy as np
from waffle_utils.log import datetime_now

from waffle_hub.schema.base_schema import BaseSchema
from waffle_hub.schema.fields import Category


@dataclass
class DatasetInfo(BaseSchema):
    name: str
    task: str
    categories: list[Category] = None
    created: str = None

    def __post_init__(self):
        self.created = self.created or datetime_now()


@dataclass
class ImageInfo(BaseSchema):
    """ImageInfo

    image_path: Image path
    ori_shape: Original image shape (Width, Height)
    new_shape: Resized image shape without padding (Width, Height)
    input_shape: Resized image shape with padding (Width, Height)
    pad: Padding (Left, Top)

    Returns:
        ImageInfo: ImageInfo
    """

    ori_shape: list[int]
    new_shape: list[int]
    input_shape: list[int]
    pad: list[int]
    ori_image: np.ndarray = None
    image_path: str = None
    image_rel_path: str = None



from abc import ABC, abstractmethod

from waffle_utils.file import io

from waffle_hub import TaskType


class BaseField(ABC):
    def __init__(self):
        pass

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({', '.join([f'{k}={v}' for k, v in self.to_dict().items()])})"

    def __str__(self) -> str:
        return self.__repr__()

    @abstractmethod
    def to_dict(self):
        pass

    @classmethod
    @abstractmethod
    def new(cls):
        pass

    @classmethod
    def from_dict(cls, d: dict, task: str = None) -> "BaseField":

        if task == TaskType.CLASSIFICATION:
            return cls.classification(**d)
        elif task == TaskType.OBJECT_DETECTION:
            return cls.object_detection(**d)
        elif task == TaskType.SEMANTIC_SEGMENTATION:
            return cls.semantic_segmentation(**d)
        elif task == TaskType.INSTANCE_SEGMENTATION:
            return cls.instance_segmentation(**d)
        elif task == TaskType.KEYPOINT_DETECTION:
            return cls.keypoint_detection(**d)
        elif task == TaskType.TEXT_RECOGNITION:
            return cls.text_recognition(**d)
        elif task == TaskType.REGRESSION:
            return cls.regression(**d)
        else:
            return cls.new(**d)

    @classmethod
    def from_json(cls, f: str, task: str = None) -> "BaseField":
        """Load Field from json file.

        Args:
            f (str): json file path.
            task (str, optional): task name. Default to None.

        Returns:
            Field Object: Field Object.
        """
        d: dict = io.load_json(f)
        return cls.from_dict(d, task)

    def __getitem__(self, key):
        return getattr(self, key)



import cv2
import numpy as np
from pycocotools import mask as mask_utils


def convert_rle_to_mask(rle: dict) -> np.ndarray:
    mask = mask_utils.frPyObjects(rle, rle["size"][0], rle["size"][1])  # height, width
    return mask_utils.decode(mask)


def convert_rle_to_polygon(rle: dict) -> list:
    mask = convert_rle_to_mask(rle)
    return convert_mask_to_polygon(mask)


def convert_mask_to_polygon(mask: np.ndarray) -> list[list]:
    contours, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    polygon = []

    for contour in contours:
        # contour = np.flip(contour, axis=1)
        # polygon.append(contour.ravel().tolist())

        p = []
        contour = contour.reshape(-1, 2)
        for point in contour:
            p.extend(point.tolist())
        polygon.append(p)
    return polygon


def merge_multi_segment(segments: list[list], image_size: tuple) -> list:
    """Merge multi segments to one list.
    Find the coordinates with min distance between each segment,
    then connect these coordinates with one thin line to merge all
    segments into one.
    Args:
        segments(List(List)): original segmentations in coco's json file.
            like [segmentation1, segmentation2,...],
            each segmentation is a list of coordinates.
    """
    if len(segments) == 0:
        return []
    if len(segments) == 1:
        return segments[0]

    # TODO: improve this function
    # find the coordinates with min distance between each segment
    # then connect these coordinates with one thin line to merge all
    # segments into one

    res = []
    for segment in segments:
        res.extend(segment)
    return res



import subprocess
from pathlib import Path

import pytest
from waffle_utils.file.io import unzip
from waffle_utils.file.network import get_file_from_url


def run_cli(cmd):
    ret = subprocess.run(cmd, check=True, shell=True)
    return ret


@pytest.fixture(scope="session")
def test_dir(tmpdir_factory):
    return Path(tmpdir_factory.mktemp("test"))


def test_dataset_new(test_dir: Path):
    cmd = f"python -m waffle_hub.dataset.cli new \
        --name new \
        --root-dir {test_dir / 'datasets'} \
        --task classification \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0


def test_dataset_from_coco(test_dir: Path):
    url = "https://raw.githubusercontent.com/snuailab/assets/main/waffle/sample_dataset/mnist.zip"
    coco_dir = test_dir / "datasets" / "mnist_coco"

    get_file_from_url(url, str(test_dir), True)
    unzip(str(test_dir / "mnist.zip"), coco_dir)

    cmd = f"python -m waffle_hub.dataset.cli from_coco \
        --name from_coco \
        --root-dir {test_dir / 'datasets'} \
        --coco-file {coco_dir / 'coco.json'} \
        --coco-root-dir {coco_dir / 'images'} \
        --task classification \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert (test_dir / "datasets" / "from_coco").exists()


def test_dataset_from_yolo(test_dir: Path):
    url = "https://raw.githubusercontent.com/snuailab/assets/main/waffle/sample_dataset/mnist_yolo_object_detection.zip"
    yolo_dir = test_dir / "datasets" / "mnist_yolo"

    get_file_from_url(url, str(test_dir), True)
    unzip(str(test_dir / "mnist_yolo_object_detection.zip"), yolo_dir)

    cmd = f"python -m waffle_hub.dataset.cli from_yolo \
        --name from_yolo \
        --root-dir {test_dir / 'datasets'} \
        --task object_detection \
        --yolo-root-dir {yolo_dir} \
        --yaml-path {yolo_dir / 'data.yaml'} \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert (test_dir / "datasets" / "from_yolo").exists()


def test_dataset_from_transformers(test_dir: Path):
    url = "https://raw.githubusercontent.com/snuailab/assets/main/waffle/sample_dataset/mnist_huggingface_classification.zip"
    hf_dir = test_dir / "datasets" / "mnist_hf"

    get_file_from_url(url, str(test_dir), True)
    unzip(str(test_dir / "mnist_huggingface_classification.zip"), hf_dir)

    cmd = f"python -m waffle_hub.dataset.cli from_transformers \
        --name from_hf \
        --root-dir {test_dir / 'datasets'} \
        --task classification \
        --dataset-dir {hf_dir} \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert (test_dir / "datasets" / "from_hf").exists()


def test_dataset_split(test_dir: Path):
    cmd = f"python -m waffle_hub.dataset.cli split \
        --name from_coco \
        --root-dir {test_dir / 'datasets'} \
        --train-ratio 0.8 \
        --val-ratio 0.1 \
        --test-ratio 0.1 \
        --method random \
        --seed 42 \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert (test_dir / "datasets" / "from_coco" / "sets" / "train.json").exists()


def test_dataset_export(test_dir: Path):
    cmd = f"python -m waffle_hub.dataset.cli export \
        --data-type ultralytics \
        --name from_coco \
        --root-dir {test_dir / 'datasets'} \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert (test_dir / "datasets" / "from_coco" / "exports" / "YOLO").exists()


def test_dataset_clone(test_dir: Path):
    cmd = f"python -m waffle_hub.dataset.cli clone \
        --src-name from_coco \
        --name clone \
        --src-root-dir {test_dir / 'datasets'} \
        --root-dir {test_dir / 'datasets'} \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert (test_dir / "datasets" / "clone").exists()


def test_dataset_delete(test_dir: Path):
    cmd = f"python -m waffle_hub.dataset.cli delete \
        --name clone \
        --root-dir {test_dir / 'datasets'} \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert not (test_dir / "datasets" / "clone").exists()


def test_dataset_get_split_ids(test_dir: Path):
    cmd = f"python -m waffle_hub.dataset.cli get_split_ids \
        --name from_coco \
        --root-dir {test_dir / 'datasets'} \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0


def test_dataset_merge(test_dir: Path):
    cmd = f"python -m waffle_hub.dataset.cli merge \
        --name merge \
        --root-dir {test_dir / 'datasets'} \
        --src-names [from_coco,from_hf] \
        --src-root-dirs {test_dir / 'datasets'} --src-root-dirs {test_dir / 'datasets'} \
        --task classification \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert (test_dir / "datasets" / "merge").exists()


def test_dataset_sample(test_dir: Path):
    cmd = f"python -m waffle_hub.dataset.cli sample \
        --name sample \
        --root-dir {test_dir / 'datasets'} \
        --task object_detection \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert (test_dir / "datasets" / "sample").exists()


def test_dataset_get_fields(test_dir: Path):
    # image
    cmd = f"python -m waffle_hub.dataset.cli get_images \
        --name sample \
        --root-dir {test_dir / 'datasets'} \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0

    # annotation
    cmd = f"python -m waffle_hub.dataset.cli get_annotations \
        --name sample \
        --root-dir {test_dir / 'datasets'} \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0

    # category
    cmd = f"python -m waffle_hub.dataset.cli get_categories \
        --name sample \
        --root-dir {test_dir / 'datasets'} \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0


def test_hub_new(test_dir: Path):
    cmd = f'python -m waffle_hub.hub.cli new \
        --backend ultralytics \
        --root-dir {test_dir / "hubs"} \
        --name test \
        --task classification \
        --model-type yolov8 \
        --model-size n \
        --categories [1,2] \
    '
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert (test_dir / "hubs" / "test").exists()


def test_hub_train(test_dir: Path):
    cmd = f'python -m waffle_hub.hub.cli train \
        --root-dir {test_dir / "hubs"} \
        --name test \
        --dataset {test_dir / "datasets" / "from_coco"} \
        --epochs 1 \
        --batch-size 4 \
        --image-size 16 \
        --learning-rate 0.001 \
        --letter-box \
        --device cpu \
        --workers 0 \
        --seed 0 \
        --verbose \
    '
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert (test_dir / "hubs" / "test" / "artifacts").exists()


def test_hub_train_advance_params(test_dir: Path):
    cmd = f'python -m waffle_hub.hub.cli new \
        --backend ultralytics \
        --root-dir {test_dir / "hubs"} \
        --name test_adv \
        --task classification \
        --model-type yolov8 \
        --model-size n \
        --categories [1,2] \
    '
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert (test_dir / "hubs" / "test").exists()

    cmd = (
        f'python -m waffle_hub.hub.cli train \
        --root-dir {test_dir / "hubs"} \
        --name test_adv \
        --dataset {test_dir / "datasets" / "from_coco"} \
        --epochs 1 \
        --batch-size 4 \
        --image-size 16 \
        --learning-rate 0.001 \
        --letter-box \
        --device cpu \
        --workers 0 \
        --seed 0 \
        --verbose \
    '
        + ' --advance_params "{box: 3}"'
    )
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert (test_dir / "hubs" / "test" / "artifacts").exists()


def test_hub_inference(test_dir: Path):
    cmd = f'python -m waffle_hub.hub.cli inference \
        --root-dir {test_dir / "hubs"} \
        --name test \
        --source {test_dir / "datasets" / "from_coco" / "raw"} \
        --confidence-threshold 0.25 \
        --device cpu \
        --workers 0 \
    '
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert (test_dir / "hubs" / "test" / "inferences").exists()


def test_hub_evaluate(test_dir: Path):
    cmd = f'python -m waffle_hub.hub.cli evaluate \
        --name test \
        --root-dir {test_dir / "hubs"} \
        --dataset {test_dir / "datasets" / "from_coco"} \
        --set-name test \
        --device cpu \
        --workers 0 \
    '
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert (test_dir / "hubs" / "test" / "evaluate.json").exists()


def test_hub_export(test_dir: Path):
    cmd = f'python -m waffle_hub.hub.cli export \
        --name test \
        --root-dir {test_dir / "hubs"} \
        --device cpu \
    '
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert (test_dir / "hubs" / "test" / "weights" / "model.onnx").exists()
